"""
########################################################################
The soil_water.py module contains the SoilWaterSeries class, which
provides I/O tools for processing measured volumetric soil water content
(SWC, cm3/cm3) data in the pyfao56 environment. The SoilWaterSeries
class manages SWC data collected at one location over time (e.g., at one
access tube over a growing season). A subclass called SoilWaterProfile
handles data storage and computations for one soil water profile
measurement event (i.e., measurements of the soil water profile on a
single date). The SoilWaterProfile class computes root zone soil water
metrics, especially the root zone soil water depletion (SWD, mm), given
estimates of root depth.

The soil_water.py module contains the following:
    SoilWaterSeries - A class for managing measured SWC profile series
    SoilWaterProfile - A class for managing a single SWC profile

10/17/2022 SWC Python functions developed by Josh Brekel, USDA-ARS
03/07/2023 SoilWater functions developed by Josh Brekel, USDA-ARS
08/23/2023 Major overhaul for pyfao56 1.2 release
########################################################################
"""

import pandas as pd
import datetime

class SoilWaterSeries:
    """A class for managing a series of measured soil water content data

    Attributes
    ----------
    par : pyfao56 Parameters object, optional
        Provides the parameter data (e.g., thetaFC, Zrmax)
        (default = None)
    sol : pyfao56 SoilProfile object, optional
        Provides layered soil profile data (e.g., thetaFC)
        (default = None)
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    swdata : dict
        Container for SoilWaterProfile objects
        key - measurement date as string ('yyyy-ddd')
        value - SoilWaterProfile object

    Methods
    -------
    savefile(filepath='pyfao56.sws')
        Save soil water series data to a file
    loadfile(filepath='pyfao56.sws')
        Load soil water series data from a file
    addprofile(mdate,swp)
        Add a SoilWaterProfile object to self.swdata
    updatepar(par)
        Update Parameter object assigned to SoilWaterProfile objects
    updatesol(sol)
        Update SoilProfile object assigned to SoilWaterProfile objects
    customload()
        Override this function to customize loading measured volumetric
        soil water content data.
    summarize()
        Summarize the series of root zone soil water metrics
    """

    def __init__(self,filepath=None,par=None,sol=None,comment=''):
        """Initialize the SoilWaterSeries class attributes.

        If filepath is provided, soil water data is loaded from the file

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None)
        par : pyfao56 Parameters object, optional
            Provides the parameter data (e.g., thetaFC, Zrmax)
            (default = None)
        sol : pyfao56 SoilProfile object, optional
            Provides layered soil profile data (e.g., thetaFC)
            (default = None)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.par = par
        self.sol = sol
        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.swdata = {}

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the SoilWaterSeries class data as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'Measured Soil Water Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
          ).format(ast,timestamp,ast,self.comment,ast)
        if len(self.swdata) > 0:
            s += 'Year-DOY  n'
            key0 = list(self.swdata.keys())[0]
            n = len(self.swdata[key0].mvswc)
            for i in range(n):
                s += ' D{:02d}'.format(i+1)
            for i in range(n):
                s += ' SWC{:02d}'.format(i+1)
            s += '    Zr      mDr   mDrmax   mfDr mfDrmax mSWCr mSWCrmax'
            s += '    mKs\n'
            for key in sorted(self.swdata.keys()):
                s += self.swdata[key].__str__() + '\n'
        return s

    def savefile(self,filepath='pyfao56.sws'):
        """Save pyfao56 soil water series data to a file.

        The function saves a file with standard pyfao56-styled header.
        Each data line contains the following info:
        1. SWC measurement date as string ('yyyy-ddd')
        2. The number, x, of SWC measurement depths as integer
        3. x bottom depths of SWC measurement layers as integer (cm)
        4. x SWC measurements as integer (cm3/cm3, same order as depths)
        5. Estimate of Zr on the measurement date
        6. mDr (mm)
        7. mDrmax (mm)
        8. mfDr (mm/mm)
        9. mfDrmax (mm/mm)
        10. mSWCr (cm3/cm3)
        11. mSWCrmax (cm3/cm3)
        12. mKs (0.0-1.0)

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sws')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for soil water data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.sws'):
        """Load measured soil water content data from a file

        The function expects a file with standard pyfao56-styled header.
        Each data line contains the following info:
        1. SWC measurement date as string ('yyyy-ddd')
        2. The number, x, of SWC measurement depths as integer
        3. x bottom depths of SWC measurement layers as integer (cm)
        4. x SWC measurements as integer (cm3/cm3, same order as depths)
        5. Optionally, an estimate of Zr on the measurement date
        6. Optionally, other values, which are not loaded.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sws')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for soil water data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1]
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            self.swdata.clear()
            for line in lines[endast+2:]:
                line = line.strip().split()
                mdate = line[0]
                numdpths = int(line[1])
                mvswc = dict()
                for i in list(range(numdpths)):
                    dpth = int(line[2+i])
                    swc = float(line[2+i+numdpths])
                    mvswc.update({dpth:swc})
                try:
                    Zr = float(line[2+numdpths*2+1])
                except:
                    Zr = float('NaN')
                swp = self.SoilWaterProfile(mdate,
                                            mvswc,
                                            par = self.par,
                                            sol = self.sol,
                                            Zr = Zr)
                self.addprofile(mdate,swp)

    def addprofile(self,mdate,swp):
        """Add a SoilWaterProfile object to self.swdata

        Parameters
        ----------
        mdate : str
            Measurement date as string ('yyyy-ddd')
        swp : SoilWaterProfile object
            Measured soil water profile data on mdate
        """

        self.swdata.update({mdate:swp})

    def updatepar(self,par):
        """Update Parameter object assigned to SoilWaterProfile objects

        Parameters
        ----------
        par : pyfao56 Parameters object
            Provides the parameter data (e.g., thetaFC, Zrmax)
        """

        self.par = par
        for mdate in self.swdata.keys():
            self.swdata[mdate].par = par

    def updatesol(self,sol):
        """Update SoilProfile object for all SoilWaterProfile objects

        Parameters
        ----------
        sol : pyfao56 SoilProfile object
            Provides layered soil profile data (e.g., thetaFC)
        """

        self.sol = sol
        for mdate in self.swdata.keys():
            self.swdata[mdate].sol = sol

    def customload(self):
        """Override this function to customize loading measured
        vol. soil water content data into SoilWaterProfile objects.
        """

        pass

    def summarize(self):
        """Summarize the series of root zone soil water metrics

        Returns
        -------
        summary : DataFrame
            Summary of series of root zone soil water metrics
        """

        cols = ['Year-DOY','mDr','mDrmax','mfDr','mfDrmax',
                'mSWCr','mSWCrmax','mKs']
        summary = pd.DataFrame(columns = cols)
        summary = summary.set_index('Year-DOY')
        for key in sorted(self.swdata.keys()):
            summary.loc[key,'mDr'] = self.swdata[key].mDr
            summary.loc[key,'mDrmax'] = self.swdata[key].mDrmax
            summary.loc[key,'mfDr'] = self.swdata[key].mfDr
            summary.loc[key,'mfDrmax'] = self.swdata[key].mfDrmax
            summary.loc[key,'mSWCr'] = self.swdata[key].mSWCr
            summary.loc[key,'mSWCrmax'] = self.swdata[key].mSWCrmax
            summary.loc[key,'mKs'] = self.swdata[key].mKs
        return summary

    class SoilWaterProfile:
        """Manage a single soil water content measurement profile

        Attributes
        ----------

        mdate : str
            Year and day of year of the measurement ('yyyy-ddd')
        mvswc : dict
            Measured volumetric soil water content (SWC, cm3/cm3)
            key - Bottom depth of measurement layer as integer (cm)
            value - Measured volumetric SWC as float (cm3/cm3)
        par : pyfao56 Parameters class, optional
            Provides the parameter data (e.g., thetaFC, Zrmax)
            (default = None)
        sol : pyfao56 SoilProfile class, optional
            Provides layered soil profile data (e.g., thetaFC)
            (default = None)
        Zr : float, optional
            Root depth on SWC measurement date (m) (default = NaN)
        mDr : float
            Measured soil water depletion for root zone (mm)
        mDrmax : float
            Measured soil water depletion for maximum root zone (mm)
        mfDr : float
            Measured fractional depletion for root zone (mm/mm)
        mfDrmax : float
            Measured fractional depletion for maximum root zone (mm/mm)
        mSWCr : float
            Measured soil water content for root zone (cm3/cm3)
        mSWCrmax : float
            Measured soil water content for maximum root zone (cm3/cm3)
        mKs : float
            Measured transpiration reduction factor (Ks, 0.0-1.0)

        Methods
        -------
        getZr(model)
            Get Zr on measurement date from model simulation output
        computeDr(negdep=True)
            Compute root zone soil water status metrics from mvswc
        computeKs(model)
            Estimate Ks from measured Dr, TAW, and RAW
        """

        def __init__(self, mdate, mvswc, par = None, sol = None,
                     Zr=float('NaN')):
            """
            Initialize the SoilWaterProfile class attributes.

            Parameters
            ----------
            mdate : str
                Year and day of year of measurement ('yyyy-ddd')
            mvswc : dict
                Measured volumetric soil water content (SWC, cm3/cm3)
                key - Bottom depth of measurement layer as integer (cm)
                value - Measured volumetric SWC as float (cm3/cm3)
            par : pyfao56 Parameters class, optional
                Provides the parameter data (e.g., thetaFC, Zrmax)
                (default = None)
            sol : pyfao56 SoilProfile class, optional
                Provides layered soil profile data (e.g., thetaFC)
                (default = None)
            Zr : float, optional
                Root depth on SWC measurement date (m) (default = NaN)
            """

            self.mdate = mdate
            self.mvswc = mvswc
            self.par = par
            self.sol = sol
            self.Zr = Zr
            self.mDr = float('NaN')
            self.mDrmax = float('NaN')
            self.mfDr = float('NaN')
            self.mfDrmax = float('NaN')
            self.mSWCr = float('NaN')
            self.mSWCrmax = float('NaN')
            self.mKs = float('NaN')

        def __str__(self):
            """Represent the SoilWaterProfile class as a string"""

            s = ('{:8s} '
                 '{:2d} '
                ).format(self.mdate,len(self.mvswc.keys()))
            for key in sorted(self.mvswc.keys()):
                s += '{:3d} '.format(key)
            for key in sorted(self.mvswc.keys()):
                s += '{:5.3f} '.format(self.mvswc[key])
            s += ('{:5.3f} {:8.3f} {:8.3f} {:6.3f} {:7.3f} '
                  '{:5.3f} {:8.3f} {:6.3f}'
                 ).format(self.Zr,self.mDr,self.mDrmax,self.mfDr,
                          self.mfDrmax,self.mSWCr,self.mSWCrmax,
                          self.mKs)
            return s

        def getZr(self, mdl):
            """Get Zr from model simulation on the measurement date

            Parameters
            ----------
            mdl : pyfao56 Model object
                Provides a Model instance with an odata DataFrame
            """

            self.Zr = mdl.odata.loc[self.mdate,'Zr']

        def computeDr(self, negdep = True):
            """Compute root zone soil water status metrics

            Parameters
            ----------
            negdep : boolean, optional
                Allow negative depletion or not (default = True)
            """

            #Root zone is evaluated in 10^-5 meter increments
            #Set root zone depth variables in 10^-5 meter units
            rzmax = int(self.par.Zrmax * 100000.) #10^-5 meters
            rz = int(self.Zr * 100000.) #10^-5 meters

            #Initialize other variables
            swc_dpths = list(self.mvswc.keys())
            if swc_dpths[-1]*1000 < rzmax:
                raise Exception("SWC measured depth must be >= Zrmax")
            if self.sol is not None:
                sol_dpths = list(self.sol.sdata.index.values) #cm
                thetaFC = self.sol.sdata['thetaFC'].to_dict()
                thetaWP = self.sol.sdata['thetaWP'].to_dict()
                if sol_dpths[-1]*1000 < rzmax:
                    raise Exception("Profile depth must be >= Zrmax")
            elif self.par is not None:
                sol_dpth = int(self.par.Zrmax*100.) #cm
                sol_dpths = [sol_dpth] #cm
                thetaFC = {sol_dpth:self.par.thetaFC}
                thetaWP = {sol_dpth:self.par.thetaWP}
            else:
                raise Exception("No soil profile data available.")
            FCr = 0.
            FCrmax = 0.
            WPr = 0.
            WPrmax = 0.
            SWCr = 0.
            SWCrmax = 0.

            #Iterate the max root zone depth in 10^-5 m increments
            inc_dpth = 0.01 #mm
            for inc in list(range(1, rzmax + 1)):
                #Find soil profile layer depth that contains inc
                sol_dpth = [dpth for (idx, dpth) in enumerate(sol_dpths)
                            if inc <= dpth * 1000][0] #10^-5 meters
                #Find SWC measurement bottom depth that contains inc
                swc_dpth = [dpth for (idx, dpth) in enumerate(swc_dpths)
                            if inc <= dpth * 1000][0] #10^-5 meters
                #Compute incremental values
                FCinc  = thetaFC[sol_dpth] * inc_dpth #mm
                WPinc  = thetaWP[sol_dpth] * inc_dpth #mm
                SWCinc = self.mvswc[swc_dpth] * inc_dpth #mm
                if not negdep and SWCinc > FCinc:
                    SWCinc = FCinc #no negative depletion

                #Accumulate
                FCrmax  += FCinc #mm
                WPrmax  += WPinc #mm
                SWCrmax += SWCinc #mm
                if inc < rz:
                    FCr  += FCinc #mm
                    WPr  += WPinc #mm
                    SWCr += SWCinc #mm

            #Finalize water status metrics
            self.mDr = FCr - SWCr #mm
            self.mDrmax = FCrmax - SWCrmax #mm
            self.mfDr = (FCr - SWCr) / (FCr - WPr) #mm/mm
            self.mfDrmax = (FCrmax-SWCrmax)/(FCrmax-WPrmax) #mm/mm
            self.mSWCr = SWCr / (rz * inc_dpth) #cm3/cm3
            self.mSWCrmax = SWCrmax / (rzmax * inc_dpth) #cm3/cm3

        def computeKs(self, mdl):
            """Estimate Ks from measured Dr, TAW, and RAW

            Parameters
            ----------
            mdl : pyfao56 Model object
                Provides a Model instance with an odata DataFrame
            """

            TAW = mdl.odata.loc[self.mdate,'TAW']
            RAW = mdl.odata.loc[self.mdate,'RAW']
            Ks = (TAW - self.mDr) / (TAW - RAW) #FAO-56 Eq. 84
            self.mKs = sorted([0.0,Ks,1.0])[1]



"""
########################################################################
The update.py module contains the Update class, which provides I/O tools
for state variable updating during FAO-56 calculations.

The update.py module contains the following:
    Update - A class for managing state variable updates with FAO-56
        calculations

11/17/2021 Finalized updates for inclusion in the pyfao56 Python package
########################################################################
"""

import pandas as pd
import datetime

class Update:
    """A class for managing update data for FAO-56 calculations.

    At this time, update methods are available for the basal crop
    coefficient (Kcb), the plant height (h), and the crop cover (fc).

    Attributes
    ----------
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    udata : DataFrame
        Update data as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Kcb','h','fc']
            Kcb - Basal crop coefficient (Kcb)
            h   - Plant height (h, m)
            fc  - Crop cover (fc, m)

    Methods
    -------
    savefile(filepath='pyfao56.upd')
        Save the update data to a file
    loadfile(filepath='pyfao56.upd')
        Load the update data from a file
    customload()
        Users can override for custom loading of update data
    getdata(index,var)
        Return a value from self.udata for model updating
    """

    def __init__(self,filepath=None,comment=''):
        """Initialize the Update class attributes.

        If filepath is provided, update data is loaded from the file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.udata = pd.DataFrame(columns=['Kcb','h','fc'])

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the Update class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        pd.options.display.float_format = '{:6.4f}'.format
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'Update Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
           'Year-DOY    Kcb      h     fc\n'
          ).format(ast,timestamp,ast,self.comment,ast)
        if not self.udata.empty:
            s += self.udata.to_string(header=False, na_rep='   NaN')
        return s

    def savefile(self,filepath='pyfao56.upd'):
        """Save pyfao56 update data to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.upd')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for update data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.upd'):
        """Load pyfao56 update data from a file.

        Parameters
        ----------
        filepath : str, optional
           Any valid filepath string (default = 'pyfao56.upd')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for update data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1] 
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            self.udata = pd.DataFrame(columns=['Kcb','h','fc'])
            for line in lines[endast+2:]:
                line = line.strip().split()
                year = line[0][:4]
                doy = line[0][-3:]
                key = '{:04d}-{:03d}'.format(int(year),int(doy))
                self.udata.loc[key] = [float(line[1]),float(line[2]),
                                       float(line[3])]

    def customload(self):
        """Override this function to customize loading update data."""

        pass

    def getdata(self, index, var):
        """Obtain an update value for a variable, if available.

        Parameters
        ----------
        index : str
            A year-doy string (yyyy-ddd) as an index to self.udata
        var : str
            A column variable to return ('Kcb','h','fc')
        """

        try:
            return self.udata.loc[index,var]
        except:
            return float('NaN')



"""
########################################################################
The example_soil.py module contains the ExampleSoil class, which
inherits from the pyfao56 SoilProfile class in soil_profile.py.
ExampleSoil provides I/O functionality for creating a SoilProfile sdata
class object from four ordered lists.

09/21/2022 Initial Python functions developed by Josh Brekel
11/30/2022 Finalized updates for inclusion in the pyfao56 Python package
########################################################################
"""

from pyfao56 import SoilProfile
import pandas as pd

class ExampleSoil(SoilProfile):
    """A class for loading soil profile data from four ordered lists.

    Inherits attributes and methods from pyfao56 SoilProfile. Overrides
    the SoilProfile customload function to allow soil profile data to be
    loaded from four ordered lists.

    Attributes
    ----------
    cnames : list
        Column names for sdata
    sdata : DataFrame
        Soil profile data as float
        index = Bottom depth of the layer as integer (cm)
        columns = ['thetaFC', 'thetaWP', 'theta0']
            thetaFC - Volumetric Soil Water Content, Field Capacity
                      (cm3/cm3)
            thetaWP - Volumetric Soil Water Content, Wilting Point
                      (cm3/cm3)
            theta0  - Volumetric Soil Water Content, Initial
                      (cm3/cm3)

    Methods
    -------
    customload()
        Overridden method from the pyfao56 SoilProfile class to provide
        customization for loading SoilProfile data from four lists.
    """

    def customload(self, bottom_depths, fc, wp, ini):
        """Loads soil profile information from four ordered lists.

        Parameters
        ----------
        bottom_depths : list
            A list of the bottom depths (cm, int) of assumed soil
            layers. Order the list from the shallowest layer to the
            deepest layer.
        fc : list
            A list of field capacities (cm3/cm3) of assumed soil
            layers. Order the list from the shallowest layer to the
            deepest layer.
        wp : list
            A list of wilting points (cm3/cm3) of assumed soil layers.
            Order the list from the shallowest layer to the deepest
            layer.
        ini : list
            A list of initial volumetric soil water content (cm3/cm3)
            measurements of assumed soil layers. Order the list from the
            shallowest layer to the deepest layer.

        ***Note on parameters: all lists should be created such that
        data is ordered from the top (shallow part) of the soil profile
        to the bottom (deep part) of the soil profile. For example, the
        FieldCapacity of the third layer down should be the third entry
        in the fc list.
        """

        soil_data_dict = {self.cnames[0]: fc,
                          self.cnames[1]: wp,
                          self.cnames[2]: ini}
        self.sdata = pd.DataFrame.from_dict(soil_data_dict,
                                            dtype='float64')
        self.sdata.index = bottom_depths
        self.sdata.index.astype('int32', copy=False)



"""
########################################################################
The custom subpackage within the pyfao56 Python package demonstrates
user customizations, particularly for loading weather and soil data.

The custom subpackage contains the following modules:
    azmet_maricopa.py - Prepares weather data from the AZMET weather
        station in Maricopa, Arizona.
    example_soil.py - provides I/O functionality for creating a
        SoilProfile sdata class object from four ordered lists

11/04/2021 Module created by Kelly Thorp
11/30/2022 example_soil.py created by Josh Brekel
########################################################################
"""

from .azmet_maricopa import AzmetMaricopa
from .example_soil import ExampleSoil



"""
########################################################################
The irrigation.py module contains the Irrigation class, which provides
I/O tools for defining the irrigation management schedule, as required
for FAO-56 calculations.

The irrigation.py module contains the following:
    Irrigation - A class for managing irrigation data for FAO-56
        calculations

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
12/13/2023 Added irrigation efficiency term for each irrigation event
02/14/2024 Added function to return date of the latest irrigation record
########################################################################
"""

import pandas as pd
import datetime

class Irrigation:
    """A class for managing irrigation data for FAO-56 calculations

    Attributes
    ----------
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    idata : DataFrame
        Irrigation data as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Depth','fw','ieff']
            Depth - Irrigation depth (mm)
            fw - fraction of soil surface wetted (FAO-56 Table 20)
            ieff - Irrigation application efficiency (%)

    Methods
    -------
    savefile(filepath='pyfao56.irr')
        Save irrigation data to a file
    loadfile(filepath='pyfao56.irr')
        Load irrigation data from a file
    addevent(year,doy,depth,fw,ieff)
        Add an irrigation event to self.idata
    customload()
        Users can override for custom loading of irrigation data.
    getlastdate()
        Return the last irrigation date in the record.
    """

    def __init__(self,filepath=None,comment=''):
        """Initialize the Irrigation class attributes.

        If filepath is provided, irrigation data is loaded from the file

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.idata = pd.DataFrame(columns=['Depth','fw','ieff'])

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the Irrigation class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        fmts = {'Depth':'{:6.2f}'.format,'fw':'{:6.2f}'.format,
                'ieff':'{:6.1f}'.format}
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'Irrigation Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
           'Year-DOY  Depth     fw IrrEff\n'
          ).format(ast,timestamp,ast,self.comment,ast)
        if not self.idata.empty:
            s += self.idata.to_string(header=False,formatters=fmts)
        return s

    def savefile(self,filepath='pyfao56.irr'):
        """Save pyfao56 irrigation data to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.irr')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for irrigation data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.irr'):
        """Load pyfao56 irrigation data from a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.irr')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for irrigation data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1]
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            self.idata = pd.DataFrame(columns=['Depth','fw','ieff'])
            for line in lines[endast+2:]:
                line = line.strip().split()
                year = line[0][:4]
                doy = line[0][-3:]
                key = '{:04d}-{:03d}'.format(int(year),int(doy))
                data = list()
                data.append(float(line[1]))
                data.append(float(line[2]))
                if len(line) == 3: #v1.2.1 and prior - no efficiency
                    data.append(100.0)
                else:
                    data.append(float(line[3]))
                self.idata.loc[key] = data

    def addevent(self, year, doy, depth, fw, ieff=100.0):
        """Add an irrigation event to self.idata

        Parameters
        ----------
        year  : int
            The 4-digit year of the irrigation event
        doy   : int
            The day of year of the irrigation event
        depth : float
            The depth of irrigation (mm)
        fw    : float
            Fraction of soil surface wetted (FAO-56 Table 20)
        ieff  : float, optional
            Irrigation application efficiency (%)
            (default = 100.0)
        """

        key = '{:04d}-{:03d}'.format(year,doy)
        self.idata.loc[key] = [depth,fw,ieff]

    def customload(self):
        """Override this function to customize loading irrigation
        data."""

        pass

    def getlastdate(self):
        """Return the last irrigation date in the record.

        Returns
        -------
        lastirr : datetime
            The date of the last irrigation in the record
        """

        dates = pd.to_datetime(self.idata.index,format='%Y-%j')
        return max(dates)



"""
########################################################################
The refet.py module contains functions for computing short-crop or
tall-crop reference evapotranspiration following the ASCE Standardized
Reference Evapotranspiration Equation.

The ASCE Standarized Reference Evapotranspiration Equation is
documented in the following publications:
https://ascelibrary.org/doi/book/10.1061/9780784408056

ASCE Task Committee on Standardization of Reference Evapotranspiration 
(Walter, I. A., Allen, R. G., Elliott, R., Itenfisu, D., Brown, P., 
Jensen, M. E.,Mecham, B., Howell, T. A., Snyder, R., Eching, S., 
Spofford, T., Hattendorf, M., Martin, D., Cuenca, R. H., Wright, J. L.)
, 2005. The ASCE Standardized Reference Evapotranspiration Equation.
American Society of Civil Engineers, Reston, VA.

The refet.py module contains the following:
    ascedaily - function to compute daily ASCE Standardized Reference ET
    ascehourly - function to compute hourly ASCE Std. Reference ET

01/07/2016 Initial Python script by Kelly Thorp
11/04/2021 Finalized updates for inclusion in pyfao56 Python package
08/01/2022 Added the ASCE hourly reference ET algorithm
08/03/2022 Added functionality to input vapor pressure
########################################################################
"""

import math

def ascedaily(rfcrp,z,lat,doy,israd,tmax,tmin,
              vapr=float('NaN'),tdew=float('NaN'),
              rhmax=float('NaN'),rhmin=float('NaN'),
              wndsp=float('NaN'),wndht=2.0):
    """Compute daily ASCE Standardized Reference Evapotranspiration

    Parameters
    ----------
    rfcrp : str
        'S' for the short reference crop (0.12-m grass)
        'T' for the tall reference crop (0.50-m alfalfa)
    z : float
        Weather site elevation above mean sea level (m)
    lat : float
        Latitude of the weather site (decimal degrees)
    doy : float
        Day number of the year between 1 and 366
    israd : float
        Incoming solar radiation (MJ m^-2 d^-1)
    tmax : float
        Daily maximum air temperature (deg C)
    tmin : float
        Daily minimum air temperature (deg C)
    vapr : float, optional (but recommended)
        Daily average vapor pressure (kPa) (default = NaN)
    tdew : float, optional (but recommended)
        Daily average dew point temperature (deg C) (default = NaN)
    rhmax : float, optional
        Daily maximum relative humidity (%) (default = NaN)
    rhmin : float, optional
        Daily minimum relative humidity (%) (default = NaN)
    wndsp : float, optional  (but recommended)
        Daily average wind speed (m s^-1) (default = NaN)
    wndht : float, optional (but recommended)
        Height of wind measurement above the ground (m) (default = 2.0)

    Returns
    -------
    etsz  : float
        Daily standardized reference evapotranspiration for the
        short or tall reference crop (mm)
    """

    #tavg (float) : Mean daily air temperature (deg C)
    #ASCE (2005) Eq. 2
    tavg = (tmax+tmin)/2.0

    #patm (float) : Mean atmospheric pressure at weather station (kPa)
    #ASCE (2005) Eq. 3
    patm = 101.3*((293.0-0.0065*z)/293.0)**5.26

    #psycon (float) : Psychrometric constant (kPa (deg C)^-1)
    #ASCE (2005) Eq. 4
    psycon = 0.000665*patm

    #Udelta (float) : Slope of the saturation vapor pressure
    #temperature curve (kPa (deg C)^-1)
    #ASCE (2005) Eq. 5
    Udelta = 2503.0*math.exp(17.27*tavg/(tavg+237.3))
    Udelta = Udelta/((tavg+237.3)**2.0)

    #es (float) : Saturation vapor pressure (kPa)
    #ASCE (2005) Eqs. 6 and 7
    emax = 0.6108*math.exp((17.27*tmax)/(tmax+237.3))
    emin = 0.6108*math.exp((17.27*tmin)/(tmin+237.3))
    es = (emax+emin)/2.0

    #ea (float): Actual vapor pressure (kPa) ASCE (2005) Table 3
    if not math.isnan(vapr):
        #ASCE (2005) Table 3
        ea = vapr
    elif not math.isnan(tdew):
        #ASCE (2005) Eq. 8
        ea = 0.6108*math.exp((17.27*tdew)/(tdew+237.3))
    elif not math.isnan(rhmax) and not math.isnan(rhmin):
        #ASCE (2005) Eq. 11
        ea = (emin*rhmax/100. + emax*rhmin/100.)/2.0
    elif not math.isnan(rhmax):
        #ASCE (2005) Eq. 12
        ea = emin*rhmax/100.
    elif not math.isnan(rhmin):
        #ASCE (2005) Eq. 13
        ea = emax*rhmin/100.
    else:
        #ASCE (2005) Appendix E
        tdew = tmin - 2.0
        ea = 0.6108*math.exp((17.27*tdew)/(tdew+237.3))

    #rns (float) : Net shortwave radiation (MJ m^-2 d^-1)
    #ASCE (2005) Eq. 16
    albedo = 0.23
    rns = (1.0-albedo)*israd

    #ra (float) : Extraterrestrial radiation (MJ m^-2 d^-1)
    #ASCE (2005) Eqs. 21-27
    latrad = lat*math.pi/180.0 #Eq. 22
    dr = 1.0+0.033*math.cos(2.0*math.pi/365.0*doy) #Eq. 23
    ldelta = 0.409*math.sin(2.0*math.pi/365.0*doy-1.39) #Eq. 24
    ws = math.acos(-1.0*math.tan(latrad)*math.tan(ldelta)) #Eq. 27
    ra1 = ws*math.sin(latrad)*math.sin(ldelta) #Eq. 21
    ra2 = math.cos(latrad)*math.cos(ldelta)*math.sin(ws) #Eq. 21
    ra = 24.0/math.pi*4.92*dr*(ra1+ra2) #Eq. 21

    #rso (float) : Clear sky solar radiation (MJ m^-2 d^-1)
    #ASCE (2005) Eq. 19
    rso = (0.75+2e-5*z)*ra

    #rnl (float) : Net longwave radiation (MJ m^-2 d^-1)
    #ASCE (2005) Eqs. 17 and 18
    ratio = sorted([0.3,israd/rso,1.0])[1]
    fcd = sorted([0.05,1.35*ratio-0.35,1.0])[1] #Eq. 18
    tk4 = ((tmax+273.16)**4.0+(tmin+273.16)**4.0)/2.0 #Eq. 17
    rnl = 4.901e-9*fcd*(0.34-0.14*math.sqrt(ea))*tk4 #Eq. 17

    #rn (float) : Net radiation (MJ m^-2 d^-1)
    #ASCE (2005) Eq. 15
    rn = rns-rnl

    #g (float) : Soil heat flux (MJ m^-2 d^-1)
    #ASCE (2005) Eq. 30
    g = 0.0

    #u2 (float) : Wind profile relationship (m s^-1)
    #ASCE (2005) Eq. 33 and Appendix E
    if math.isnan(wndsp): wndsp = 2.0
    u2 = wndsp * (4.87/math.log(67.8*wndht-5.42))

    #Aerodynamic roughness and surface resistance constants
    #ASCE (2005) Table 1
    if rfcrp == 'S': #Short reference crop (0.12-m grass)
        Cn = 900.0  #K mm s^3 Mg^-1 d^-1
        Cd = 0.34   #s m^-1
    elif rfcrp == 'T': #Tall reference crop (0.50-m alfalfa)
        Cn = 1600.0 #K mm s^3 Mg^-1 d^-1
        Cd = 0.38   #s m^-1

    #etsz (float) : Standardized daily reference crop ET (mm d^-1)
    #ASCE (2005) Eq. 1
    etsz = 0.408*Udelta*(rn-g)+psycon*(Cn/(tavg+273.0))*u2*(es-ea)
    etsz = etsz/(Udelta+psycon*(1.0+Cd*u2))

    return etsz

def ascehourly(rfcrp,z,lat,lon,lzn,doy,sct,israd,tavg,vapr=float('NaN'),
               tdew=float('NaN'),rhum=float('NaN'),tmin=float('NaN'),               wndsp=float('NaN'),wndht=2.0,tl=1.0,csreq='D',fcdpt=1.0):
    """Compute hourly ASCE Standardized Reference Evapotranspiration

    Parameters
    ----------
    rfcrp : str
        'S' for the short reference crop (0.12-m grass)
        'T' for the tall reference crop (0.50-m alfalfa)
    z : float
        Weather site elevation above mean sea level (m)
    lat : float
        Latitude of the weather site (decimal degrees)
    lon : float
        Longitude of the weather site (decimal degrees)
    lzn : float
        Longitude of the center of the local time zone (decimal degrees)
    doy : float
        Day number of the year between 1 and 366
    sct : float
        Standard clock time at the midpoint of the period (h)
        For the period between 1400-1500 hours, sct = 14.5 h
    israd : float
        Incoming solar radiation (MJ m^-2 d^-1)
    tavg : float
        Average air temperature (deg C)
    vapr : float, optional (but recommended)
        Average vapor pressure (kPa) (default = NaN)
    tdew : float, optional (but recommended)
        Average dew point temperature (deg C) (default = NaN)
    rhum : float, optional
        Average relative humidity (%) (default = NaN)
    tmin : float, optional
        Daily minimum air temperature (deg C) (default = NaN)
    wndsp : float, optional  (but recommended)
        Average wind speed (m s^-1) (default = NaN)
    wndht : float, optional (but recommended)
        Height of wind measurement above the ground (m) (default = 2.0)
    tl : float, optional
        Length of the calculation period (h) (default = 1.0)
    csreq : str, optional
        'S' for the simple Eq. 47 for clear sky solar radiation
        'D' for the complex method in Appendix D (default = 'D')
    fcdpt : float, optional
        Cloudiness value (fcd) from previous timestep (default = 1.0)

    Returns
    -------
    etsz : float
        Hourly standardized reference evapotranspiration for the
        short or tall reference crop (mm)
    fcd : Cloudiness value for possible use in next timestep
    """

    #patm (float) : Mean atmospheric pressure at weather station (kPa)
    #ASCE (2005) Eq. 34
    patm = 101.3*((293.0-0.0065*z)/293.0)**5.26

    #psycon (float) : Psychrometric constant (kPa (deg C)^-1)
    #ASCE (2005) Eq. 35
    psycon = 0.000665*patm

    #Udelta (float) : Slope of the saturation vapor pressure
    #temperature curve (kPa (deg C)^-1)
    #ASCE (2005) Eq. 36
    Udelta = 2503.0*math.exp(17.27*tavg/(tavg+237.3))
    Udelta = Udelta/((tavg+237.3)**2.0)

    #es (float) : Saturation vapor pressure (kPa)
    #ASCE (2005) Eq. 37
    es = 0.6108*math.exp((17.27*tavg)/(tavg+237.3))

    #ea (float): Actual vapor pressure (kPa) ASCE (2005) Table 4
    if not math.isnan(vapr):
        #ASCE (2005) Table 4
        ea = vapr
    elif not math.isnan(tdew):
        #ASCE (2005) Eq. 38
        ea = 0.6108*math.exp((17.27*tdew)/(tdew+237.3))
    elif not math.isnan(rhum):
        #ASCE (2005) Eq. 41
        ea = es*rhum/100.
    else:
        #ASCE (2005) Appendix E
        tdew = tmin - 2.0
        ea = 0.6108*math.exp((17.27*tdew)/(tdew+237.3))

    #rns (float) : Net shortwave radiation (MJ m^-2 h^-1)
    #ASCE (2005) Eq. 43
    albedo = 0.23
    rns = (1.0-albedo)*israd

    #ra (float) : Extraterrestrial radiation (MJ m^-2 h^-1)
    #ASCE (2005) Eqs. 48-58
    dr = 1.0+0.033*math.cos(2.0*math.pi/365.0*doy) #Eq. 50
    ldelta = 0.409*math.sin(2.0*math.pi/365.0*doy-1.39) #Eq. 51
    b = 2.0*math.pi*(doy-81.0)/364.0 #Eq. 58
    sc = 0.1645*math.sin(2.0*b)-0.1255*math.cos(b)-0.025*math.sin(b) #57
    wmid = math.pi/12.0*((sct+0.06667*(lzn-lon)+sc)-12.) #Eq. 55
    w1 = wmid-math.pi*tl/24.0 #Eq. 53
    w2 = wmid+math.pi*tl/24.0 #Eq. 54
    latrad = lat*math.pi/180.0 #Eq. 49
    ws = math.acos(-1.0*math.tan(latrad)*math.tan(ldelta)) #Eq. 59
    if w1 < -1.0*ws: w1 = -1.0*ws #Eq. 56
    if w2 < -1.0*ws: w2 = -1.0*ws #Eq. 56
    if w1 > ws: w1 = ws #Eq. 56
    if w2 > ws: w2 = ws #Eq. 56
    if w1 > w2: w1 = w2 #Eq. 56
    ra1 = (w2-w1)*math.sin(latrad)*math.sin(ldelta) #Eq. 48
    ra2 = math.cos(latrad)*math.cos(ldelta) #Eq. 48
    ra3 = math.sin(w2)-math.sin(w1) #Eq. 48
    if wmid < -1.0*ws or wmid > ws:
        ra = 0.0
    else:
        ra = 12.0/math.pi*4.92*dr*(ra1+ra2*ra3) #Eq. 48

    #rso (float) : Clear sky solar radiation (MJ m^-2 h^-1)
    #ASCE (2005) Eq. 47 and Appendix D
    beta1 = math.sin(latrad)*math.sin(ldelta)
    beta2 = math.cos(latrad)*math.cos(ldelta)*math.cos(wmid)
    beta = math.asin(beta1+beta2) #Eq. 62 or D.6
    if csreq == 'S':
        rso = (0.75+2e-5*z)*ra #Eq. 47
    else:
        if beta < 0.3:
            rso = 0.0
        else:
            pwat = 0.14*ea*patm+2.1 #Eq. D.3
            kt = 1.0
            kb1 = -0.00146*patm/(kt*math.sin(beta))
            kb2 = 0.075*(pwat/math.sin(beta))**0.4
            kb = 0.98*math.exp(kb1-kb2) #Eq. D.2
            if kb >= 0.15:
                kd = 0.35-0.36*kb #Eq. D.4
            else:
                kd = 0.18+0.82*kb #Eq. D.4
            rso = (kb + kd)*ra #Eq. D.1

    #rnl (float) : Net longwave radiation (MJ m^-2 h^-1)
    #ASCE (2005) Eqs. 44, 45, and 62
    if beta < 0.3 or rso <= 0.0: #nighttime
        fcd = fcdpt
    else: #daytime
        ratio = sorted([0.3,israd/rso,1.0])[1] #Eq. 45
        fcd = sorted([0.05,1.35*ratio-0.35,1.0])[1] #Eq. 45
    tk4 = (tavg+273.16)**4.0 #Eq. 44
    rnl = 2.042e-10*fcd*(0.34-0.14*math.sqrt(ea))*tk4 #Eq. 44

    #rn (float) : Net radiation (MJ m^-2 h^-1)
    #ASCE (2005) Eq. 42
    rn = rns-rnl

    #g (float) : Soil heat flux (MJ m^-2 h^-1)
    #ASCE (2005) Eqs. 65 and 66
    #Aerodynamic roughness and surface resistance constants
    #ASCE (2005) Table 1
    if rfcrp == 'S': #Short reference crop (0.12-m grass)
        Cn = 37.0 #K mm s^3 Mg^-1 h^-1
        if rn < 0.0: #nighttime
            g = 0.5*rn
            Cd = 0.96 #!s m^-1
        else: #daytime
            g = 0.1*rn
            Cd = 0.24 #!s m^-1
    elif rfcrp == 'T': #Tall reference crop (0.50-m alfalfa)
        Cn = 66.0 #K mm s^3 Mg^-1 h^-1
        if rn < 0.0: #nighttime
            g = 0.2*rn
            Cd = 1.7 #!s m^-1
        else: #daytime
            g = 0.04*rn
            Cd = 0.25 #!s m^-1

    #u2 (float) : Wind profile relationship (m s^-1)
    #ASCE (2005) Eq. 67 and Appendix E
    if math.isnan(wndsp): wndsp = 2.0
    u2 = wndsp * (4.87/math.log(67.8*wndht-5.42))

    #etsz (float) : Standardized daily reference crop ET (mm h^-1)
    #ASCE (2005) Eq. 1
    etsz = 0.408*Udelta*(rn-g)+psycon*(Cn/(tavg+273.0))*u2*(es-ea)
    etsz = etsz/(Udelta+psycon*(1.0+Cd*u2))

    return (etsz, fcd)



"""
########################################################################
The parameters.py module contains the Parameters class, which provides
I/O tools for defining input parameters, as required for FAO-56
calculations.

The parameters.py module contains the following:
    Parameters - A class for managing input parameters for FAO-56
        calculations

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
12/12/2023 Added CN2 parameter for runoff method
09/30/2024 Added inputs for single crop coefficient computation
01/27/2025 Added methods for obtaining values from FAO-56 tables
########################################################################
"""

import datetime
from pyfao56.tools import FAO56Tables

class Parameters:
    """A class for managing input parameters for FAO-56 calculations

    Attributes
    ----------
    Kcmini : float
        Single (mean) crop coefficient Initial (FAO-56 Table 12)
    Kcmmid : float
        Single (mean) crop coefficient Mid (FAO-56 Table 12)
    Kcmend : float
        Single (mean) crop coefficient End (FAO-56 Table 12)
    Kcbini : float
        Basal crop coefficient Initial (FAO-56 Table 17)
    Kcbmid : float
        Basal crop coefficient Mid (FAO-56 Table 17)
    Kcbend : float
        Basal crop coefficient End (FAO-56 Table 17)
    Lini : int
        Length Stage Initial (days) (FAO-56 Table 11)
    Ldev : int
        Length Stage Development (days) (FAO-56 Table 11)
    Lmid : int
        Length Stage Mid (days) (FAO-56 Table 11)
    Lend : int
        Length Stage End (days) (FAO-56 Table 11)
    hini : float
        Plant Height Initial (m)
    hmax : float
        Plant Height Maximum (m) (FAO-56 Table 12)
    thetaFC : float
        Volumetric Soil Water Content, Field Capacity (cm3/cm3)
    thetaWP : float
        Volumetric Soil Water Content, Wilting Point (cm3/cm3)
    theta0 : float
        Volumetric Soil Water Content, Initial (cm3/cm3)
    Zrini : float
        Rooting Depth Initial (m)
    Zrmax : float
        Rooting Depth Maximum (m) (FAO-56 Table 22)
    pbase : float
        Depletion Fraction (p) (FAO-56 Table 22)
    Ze : float
        Depth of surface evaporation layer (m) (FAO-56 Table 19 & p144)
    REW : float
        Total depth Stage 1 evaporation (mm) (FAO-56 Table 19)
    CN2 : int
        Curve Number for AWC II (ASCE (2016), Table 14-3, p452)
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class

    Methods
    -------
    savefile(filepath='pyfao56.par')
        Save the parameter data to a file
    loadfile(filepath='pyfao56.par')
        Load the parameter data from a file
    setfrom11(index, setLini=True, setLdev=True, setLmid=True,
              setLend=True):
        Set growth stage lengths from FAO-56 Table 11
    setfrom12(index, setKcmini=True, setKcmmid=True,
              setKcmend=True, sethmax=True):
        Set Kc values and maximum crop height from FAO-56 Table 12
    setfrom17(index, setKcbini=True, setKcbmid=True,
              setKcbend=True):
        Set Kcb values from FAO-56 Table 17
    setfrom22(index, setZrmax1=True, setZrmax2=True,
              setpbase=True):
        Set Zrmax and pbase values from FAO-56 Table 22
    """

    def __init__(self, Kcmini=0.35, Kcmmid=1.15, Kcmend=0.60,
                 Kcbini=0.15, Kcbmid=1.10, Kcbend=0.50, Lini=25,
                 Ldev=50, Lmid=50, Lend=25, hini=0.010, hmax=1.20,
                 thetaFC=0.250, thetaWP=0.100, theta0=0.100, Zrini=0.20,
                 Zrmax=1.40, pbase=0.50, Ze=0.10, REW=8.0, CN2=70,
                 comment=''):
        """Initialize the Parameters class attributes.

        Default parameter values are given below. Users should update
        the parameters with values for their specific crop and field
        conditions based on FAO-56 documentation.

        Parameters
        ----------
        See Parameters class docstring for parameter definitions.
        Kcmini  : float, optional, default = 0.35
        Kcmmid  : float, optional, default = 1.15
        Kcmend  : float, optional, default = 0.60
        Kcbini  : float, optional, default = 0.15
        Kcbmid  : float, optional, default = 1.10
        Kcbend  : float, optional, default = 0.50
        Lini    : int  , optional, default = 25
        Ldev    : int  , optional, default = 50
        Lmid    : int  , optional, default = 50
        Lend    : int  , optional, default = 25
        hini    : float, optional, default = 0.010
        hmax    : float, optional, default = 1.20
        thetaFC : float, optional, default = 0.250
        thetaWP : float, optional, default = 0.100
        theta0  : float, optional, default = 0.100
        Zrini   : float, optional, default = 0.20
        Zrmax   : float, optional, default = 1.40
        pbase   : float, optional, default = 0.50
        Ze      : float, optional, default = 0.10
        REW     : float, optional, default = 8.0
        CN2     : int  , optional, default = 70
        comment : str  , optional, default = ''
        """

        self.Kcmini  = Kcmini
        self.Kcmmid  = Kcmmid
        self.Kcmend  = Kcmend
        self.Kcbini  = Kcbini
        self.Kcbmid  = Kcbmid
        self.Kcbend  = Kcbend
        self.Lini    = Lini
        self.Ldev    = Ldev
        self.Lmid    = Lmid
        self.Lend    = Lend
        self.hini    = hini
        self.hmax    = hmax
        self.thetaFC = thetaFC
        self.thetaWP = thetaWP
        self.theta0  = theta0
        self.Zrini   = Zrini
        self.Zrmax   = Zrmax
        self.pbase   = pbase
        self.Ze      = Ze
        self.REW     = REW
        self.CN2     = CN2
        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp  = datetime.datetime.now()

    def __str__(self):
        """Represent the Parameter class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'Parameter Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:9.4f} Kcmini, Kc Initial (FAO-56 Table 12)\n'
           '{:9.4f} Kcmmid, Kc Mid (FAO-56 Table 12)\n'
           '{:9.4f} Kcmend, Kc End (FAO-56 Table 12)\n'
           '{:9.4f} Kcbini, Kcb Initial (FAO-56 Table 17)\n'
           '{:9.4f} Kcbmid, Kcb Mid (FAO-56 Table 17)\n'
           '{:9.4f} Kcbend, Kcb End (FAO-56 Table 17)\n'
           '{:9d} Lini, Length Stage Initial (days) (FAO-56 Table 11)\n'
           '{:9d} Ldev, Length Stage Development (days) '
           '(FAO-56 Table 11)\n'
           '{:9d} Lmid, Length Stage Mid (days) (FAO-56 Table 11)\n'
           '{:9d} Lend, Length Stage End (days) (FAO-56 Table 11)\n'
           '{:9.4f} hini, Plant Height Initial (m)\n'
           '{:9.4f} hmax, Plant Height Maximum (m) (FAO-56 Table 12)\n'
           '{:9.4f} thetaFC, Vol. Soil Water Content, Field Capacity '
           '(cm3/cm3)\n'
           '{:9.4f} thetaWP, Vol. Soil Water Content, Wilting Point '
           '(cm3/cm3)\n'
           '{:9.4f} theta0, Vol. Soil Water Content, Initial '
           '(cm3/cm3)\n'
           '{:9.4f} Zrini, Rooting Depth Initial (m)\n'
           '{:9.4f} Zrmax, Rooting Depth Maximum (m) '
           '(FAO-56 Table 22)\n'
           '{:9.4f} pbase, Depletion Fraction (p) (FAO-56 Table 22)\n'
           '{:9.4f} Ze, Depth of surface evaporation layer (m) '
           '(FAO-56 Table 19 and Page 144)\n'
           '{:9.4f} REW, Total depth Stage 1 evaporation (mm) '
           '(FAO-56 Table 19)\n'
           '{:9d} CN2, Curve Number for AWCII '
           '(ASCE (2016) Table 14-3 or SCS (1972))\n'
          ).format(ast,timestamp,ast,self.comment,ast,self.Kcmini,
                   self.Kcmmid,self.Kcmend,self.Kcbini,self.Kcbmid,
                   self.Kcbend,self.Lini,self.Ldev,self.Lmid,
                   self.Lend,self.hini,self.hmax,self.thetaFC,
                   self.thetaWP,self.theta0,self.Zrini,self.Zrmax,
                   self.pbase,self.Ze,self.REW,self.CN2)
        return s

    def savefile(self,filepath='pyfao56.par'):
        """Save pyfao56 parameters to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.par')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for parameter data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.par'):
        """Load pyfao56 parameters from a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.par')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for parameter data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1] 
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            for line in lines[endast+1:]:
                line = line.strip().split(',')[0].split()
                if line[1].lower() == 'kcmini':
                    self.Kcmini = float(line[0])
                elif line[1].lower() == 'kcmmid':
                    self.Kcmmid = float(line[0])
                elif line[1].lower() == 'kcmend':
                    self.Kcmend = float(line[0])
                elif line[1].lower() == 'kcbini':
                    self.Kcbini = float(line[0])
                elif line[1].lower() == 'kcbmid':
                    self.Kcbmid = float(line[0])
                elif line[1].lower() == 'kcbend':
                    self.Kcbend = float(line[0])
                elif line[1].lower() == 'lini':
                    self.Lini = int(line[0])
                elif line[1].lower() == 'ldev':
                    self.Ldev = int(line[0])
                elif line[1].lower() == 'lmid':
                    self.Lmid = int(line[0])
                elif line[1].lower() == 'lend':
                    self.Lend = int(line[0])
                elif line[1].lower() == 'hini':
                    self.hini = float(line[0])
                elif line[1].lower() == 'hmax':
                    self.hmax = float(line[0])
                elif line[1].lower() == 'thetafc':
                    self.thetaFC = float(line[0])
                elif line[1].lower() == 'thetawp':
                    self.thetaWP = float(line[0])
                elif line[1].lower() == 'theta0':
                    self.theta0 = float(line[0])
                elif line[1].lower() == 'zrini':
                    self.Zrini = float(line[0])
                elif line[1].lower() == 'zrmax':
                    self.Zrmax = float(line[0])
                elif line[1].lower() == 'pbase':
                    self.pbase = float(line[0])
                elif line[1].lower() == 'ze':
                    self.Ze = float(line[0])
                elif line[1].lower() == 'rew':
                    self.REW = float(line[0])
                elif line[1].lower() == 'cn2':
                    self.CN2 = int(line[0])

    def setfrom11(self, index, setLini=True, setLdev=True,
                  setLmid=True, setLend=True):
        """Set growth stage lengths from FAO-56 Table 11

        Parameters
        ----------
        index : int
            row index for FAO56Tables.table11 DataFrame
        setLini : boolean, optional
            Set Lini value from FAO-56 Table 11 or not (default = True)
        setLdev : boolean, optional
            Set Ldev value from FAO-56 Table 11 or not (default = True)
        setLmid : boolean, optional
            Set Lmid value from FAO-56 Table 11 or not (default = True)
        setLend : boolean, optional
            Set Lend value from FAO-56 Table 11 or not (default = True)
        """

        table11 = FAO56Tables().table11
        value = table11.iloc[index].loc['Lini']
        if setLini and value!='': self.Lini = int(value)
        value = table11.iloc[index].loc['Ldev']
        if setLdev and value!='': self.Ldev = int(value)
        value = table11.iloc[index].loc['Lmid']
        if setLmid and value!='': self.Lmid = int(value)
        value = table11.iloc[index].loc['Llate']
        if setLend and value!='': self.Lend = int(value)

    def setfrom12(self, index, setKcmini=True, setKcmmid=True,
                  setKcmend=True, sethmax=True):
        """Set Kcm values and maximum crop height from FAO-56 Table 12

        Parameters
        ----------
        index : int
            row index for FAO56Tables.table12 DataFrame
        setKcmini : boolean, optional
            Set Kcmini value from FAO-56 Table 12 or not (default=True)
        setKcmmid : boolean, optional
            Set Kcmmid value from FAO-56 Table 12 or not (default=True)
        setKcmend : boolean, optional
            Set Kcmend value from FAO-56 Table 12 or not (default=True)
        sethmax : boolean, optional
            Set hmax value from FAO-56 Table 12 or not (default=True)
        """

        table12 = FAO56Tables().table12
        value = table12.iloc[index].loc['Kcmini']
        if setKcmini and value!='': self.Kcmini = float(value)
        value = table12.iloc[index].loc['Kcmmid']
        if setKcmmid and value!='': self.Kcmmid = float(value)
        value = table12.iloc[index].loc['Kcmend']
        if setKcmend and value!='': self.Kcmend = float(value)
        value = table12.iloc[index].loc['hmax']
        if sethmax and value!='': self.hmax = float(value)

    def setfrom17(self, index, setKcbini=True, setKcbmid=True,
                  setKcbend=True):
        """Set Kcb values from FAO-56 Table 17

        Parameters
        ----------
        index : int
            row index for FAO56Tables.table17 DataFrame
        setKcbini : boolean, optional
            Set Kcbini value from FAO-56 Table 17 or not (default=True)
        setKcbmid : boolean, optional
            Set Kcbmid value from FAO-56 Table 17 or not (default=True)
        setKcbend : boolean, optional
            Set Kcbend value from FAO-56 Table 17 or not (default=True)
        """

        table17 = FAO56Tables().table17
        value = table17.iloc[index].loc['Kcbini']
        if setKcbini and value!='': self.Kcbini = float(value)
        value = table17.iloc[index].loc['Kcbmid']
        if setKcbmid and value!='': self.Kcbmid = float(value)
        value = table17.iloc[index].loc['Kcbend']
        if setKcbend and value!='': self.Kcbend = float(value)

    def setfrom22(self, index, setZrmax1=True, setZrmax2=True,
                  setpbase=True):
        """Set Zrmax and pbase values from FAO-56 Table 22

        Parameters
        ----------
        index : int
            row index for FAO56Tables.table22 DataFrame
        setZrmax1 : boolean, optional
            Use Zrmax1 value from FAO-56 Table 22 or not (default=True)
            If setZrmax2 is also True, use mean of Zrmax1 and Zrmax2.
        setZrmax2 : boolean, optional
            Use Zrmax2 value from FAO-56 Table 22 or not (default=True)
            If setZrmax1 is also True, use mean of Zrmax1 and Zrmax2.
        setpbase : boolean, optional
            Use pbase value from FAO-56 Table 22 or not (default=True)
        """

        table22 = FAO56Tables().table22
        if setZrmax1 and setZrmax2:
            a = float(table22.iloc[index].loc['Zrmax1'])
            b = float(table22.iloc[index].loc['Zrmax2'])
            self.Zrmax = (a + b) / 2.0
        elif setZrmax1:
            self.Zrmax = float(table22.iloc[index].loc['Zrmax1'])
        elif setZrmax2:
            self.Zrmax = float(table22.iloc[index].loc['Zrmax2'])

        if setpbase:
            self.pbase = float(table22.iloc[index].loc['pbase'])



"""
########################################################################
The blue_green.py module contains the BlueGreen class, which provides
a methodology for partitioning water balance components into blue
(irrigation) and green (rainfall) water. The class implements the
blue-green water accounting framework based on Hoekstra (2019), applying
it to FAO-56 dual crop coefficient daily soil water balance model
outputs. It tracks the fraction of soil water storage attributable to
effective precipitation (EP, green water) versus applied water (AW, blue
water) and partitions evapotranspiration, evaporation, transpiration,
runoff, and deep percolation accordingly.

Hoekstra, A. Y., 2019. Green-blue water accounting in a soil water
balance. Advances in Water Resources 129, 112-117.

The blue_green.py module contains the following:
    BlueGreen - A class for attributing output to blue & green water

11/20/2025 Initial code developed by Gulati & Kukal, University of Idaho
02/09/2026 Reformatted code for incorporation into pyfao56
########################################################################
"""

import pandas as pd
import datetime
import matplotlib.pyplot as plt

class BlueGreen:
    """
    A class for attributing pyfao56 Model output to blue and green water

    Attributes
    ----------
    mdl : pyfao56 Model class
       Provides the pyfao56 simulated data
    cg : float
       Initial fraction of effective precipitation (green water) in
       available soil water storage
    cb : float
       Initial fraction of applied water (blue water) in available
       soil water storage
    bgdata : DataFrame or None
       Result of running the blue green partitioning model as float
       index - Year and day of year as string ('yyyy-ddd')
       columns - ['Year', 'DOY', 'DOW', 'Date', 'Rain', 'Irrig', 'S',
                  'Sep', 'Saw', 'ETa', 'ETa_ep', 'ETa_aw', 'E', 'E_ep',
                  'E_aw', 'T', 'T_ep', 'T_aw', 'Runoff', 'Runoff_ep',
                  'Runoff_aw', 'DP', 'DP_ep', 'DP_aw']
           Year      - 4-digit year (yyyy)
           DOY       - Day of year (ddd)
           DOW       - Day of week
           Date      - Month/Day/Year (mm/dd/yy)
           Rain      - Depth of precipitation (mm)
           Irrig     - Depth of applied irrigation (mm)
           S         - Total available water (mm)
           Sep       - S attributed to effective precipitation (mm)
           Saw       - S attributed to applied water (mm)
           ETa       - Actual evapotranspiration (mm)
           ETa_ep    - ETa attributed to effective precipitation (mm)
           ETa_aw    - ETa attributed to applied water (mm)
           E         - Soil water evaporation (mm)
           E_ep      - E attributed to effective precipitation (mm)
           E_aw      - E attributed to applied water (mm)
           T         - Actual plant transpiration (mm)
           T_ep      - T attributed to effective precipitation (mm)
           T_aw      - T attributed to applied water (mm)
           Runoff    - Surface runoff (mm)
           Runoff_ep - Runoff attributed to effective precipitation (mm)
           Runoff_aw - Runoff attributed to applied water (mm)
           DP        - Deep percolation (mm)
           DP_ep     - DP attributed to effective precipitation (mm)
           DP_aw     - DP attributed to applied water (mm)

    Methods
    -------
    run()
        Execute the blue-green water partitioning algorithm
    savefile(filepath='bluegreen_output.iwp')
        Save the formatted blue-green partitioned results to a text file
    plot(var='ETa', filename=None)
        Plot time series of a variable and its EP and AW components
    """

    def __init__(self, mdl, cg=0.5, cb=0.5, comment=''):
        """Initialize the BlueGreen class attributes.

        Parameters
        ----------
        mdl : pyfao56 Model object
           Object with `odata` DataFrame attribute indexed as YYYY-DOY
           and containing: ['Year', 'DOY', 'Date', 'ETa', 'E', 'T', 
           'Runoff', 'DP', 'Rain', 'Irrig', 'TAW', 'Dr']
        cg : float, optional
           Initial fraction of effective precipitation (green water) in
           available soil water storage
           (default is 0.5)
        cb : float, optional
           Initial fraction of applied water (blue water) in available
           soil water storage
           (default is 0.5)
        comment : str, optional
           User-defined file descriptions or metadata (default = '')
        """
        self.df = mdl.odata.copy().iloc[:, :-4]
        self.cg = cg
        self.cb = cb
        self.comment = 'Comments: ' + comment.strip()
        self.bg_cnames = ['Year', 'DOY', 'DOW', 'Date', 'Rain', 'Irrig',
                          'S', 'Sep', 'Saw', 'ETa', 'ETa_ep', 'ETa_aw',
                          'E', 'E_ep', 'E_aw', 'T', 'T_ep', 'T_aw',
                          'Runoff', 'Runoff_ep', 'Runoff_aw', 'DP',
                          'DP_ep', 'DP_aw']
        self.startDate = mdl.startDate
        self.endDate = mdl.endDate
        self.tmstmp = datetime.datetime.now()
        self.bgdata = pd.DataFrame(columns=self.bg_cnames)

    def _water_fraction(self, S_tot_prev, S_py_prev, Win_py, Win_sy,
                        ETa, RO, DP):
        """Calculate water allocation for either the EP or AW component
           based on previous day storage

        Parameters
        ----------
        S_tot_prev : float
            Total available soil water on previous day (mm)
        S_py_prev : float
            Previous day's primary storage (EP or AW) (mm)
        Win_py : float
            Primary wetting influx (e.g., irrigation for EP)
        Win_sy : float
            Secondary wetting influx (e.g., rainfall for EP)
        ETa : float
            Actual evapotranspiration (ETa, mm)
        RO : float
            Runoff (mm)
        DP : float
            Deep percolation (mm)

        Returns
        -------
        S_py_new : float
            Updated water storage component (mm)
        """
        if S_tot_prev == 0: #Should not occur but included for safety
            S_py_new = S_py_prev + Win_py \
                - (0. if RO==0 else (Win_py / (Win_py + Win_sy)) * RO)
        else:
            S_py_new = S_py_prev + Win_py \
                - (0. if RO==0 else (Win_py / (Win_py + Win_sy)) * RO) \
                - ((S_py_prev / S_tot_prev) * ETa) \
                - ((S_py_prev / S_tot_prev) * DP)
        S_py_new = max(0. , S_py_new)
        return S_py_new

    def run(self):
        """Run blue-green partitioning model on the loaded dataframe"""

        df = self.df

        sep, saw = 0.0, 0.0  # initialize EP and AW storage

        for idx, row in df.iterrows():
            ETa = row['ETa']
            RO = row['Runoff']
            DP = row['DP']
            Rain = row['Rain']
            Irrig = row['Irrig']
            #Available soil water from previous day
            S_prev = S if idx > self.startDate.strftime('%Y-%j') else 0.
            #Available soil water from current day
            S = row['TAW'] - row['Dr']
            E = row['E']
            T = row['T']

            # To make sure S remains non-zero, use previous day's S
            if S <= 0.:
                S = S_prev

            if idx == self.startDate.strftime('%Y-%j'):
                sep = self.cg * S
                saw = self.cb * S
            else:
                # Update EP and AW storage using previous day's values
                sep = self._water_fraction(S_prev, sep, Rain, Irrig,
                                           ETa, RO, DP)
                saw = self._water_fraction(S_prev, saw, Irrig, Rain,
                                           ETa, RO, DP)

            # Save non-zero sep and saw to avoid zero divide
            if sep + saw > 0.:
                last_sep, last_saw = sep, saw
            # Handle case where both sep and saw are zero
            if sep + saw == 0.:
                sep, saw = last_sep, last_saw

            fep = sep / (sep + saw)
            faw = saw / (sep + saw)

            row_data = [
                row['Year'], row['DOY'], row['DOW'], row['Date'],
                Rain, Irrig, S, sep, saw,
                ETa, fep * ETa, faw * ETa,
                E, fep * E, faw * E,
                T, fep * T, faw * T,
                RO, fep * RO, faw * RO,
                DP, fep * DP, faw * DP
            ]

            self.bgdata.loc[idx] = row_data

    def __str__(self):
        """Represent BlueGreen water allocation results as string"""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        sdate = self.startDate.strftime('%m/%d/%Y')
        edate = self.endDate.strftime('%m/%d/%Y')

        method_note = 'Based on conceptual framework of Hoekstra (2019)'
        ast = '*' * 72

        # Format specifiers
        fmts = {
            'Year': '{:>4}'.format, 'DOY': '{:>3}'.format,
            'Date': '{:10s}'.format, 'Rain': '{:7.2f}'.format,
            'Irrig': '{:7.2f}'.format, 'S': '{:7.2f}'.format,
            'Sep': '{:7.2f}'.format, 'Saw': '{:7.2f}'.format,
            'ETa': '{:7.2f}'.format, 'ETa_ep': '{:7.2f}'.format,
            'ETa_aw': '{:7.2f}'.format, 'E': '{:7.2f}'.format,
            'E_ep': '{:7.2f}'.format, 'E_aw': '{:7.2f}'.format,
            'T': '{:7.2f}'.format, 'T_ep': '{:7.2f}'.format,
            'T_aw': '{:7.2f}'.format, 'Runoff': '{:7.2f}'.format,
            'Runoff_ep': '{:7.2f}'.format,'Runoff_aw': '{:7.2f}'.format,
            'DP': '{:7.2f}'.format, 'DP_ep': '{:7.2f}'.format, 
            'DP_aw': '{:7.2f}'.format
        }

        header_note = (
            f"{ast}\n"
            f"pyfao56 wetting influx partitioning into "
            f"applied water (AP) (blue water) and "
            f"effective precipitation (EP) (green water) output\n"
            f"Timestamp: {timestamp}\n"
            f"Simulation Start Date: {sdate}\n"
            f"Simulation End Date: {edate}\n"
            f"Partitioning Method: {method_note}\n"
            f"{ast}\n"
            f"{self.comment}\n"
            f"{ast}\n"
            "Year-DOY Year DOY DOW     Date      Rain   Irrig       S" \
            "     Sep     Saw     ETa  ETa_ep  ETa_aw       E    E_ep" \
            "    E_aw       T    T_ep    T_aw      RO   RO_ep   RO_aw" \
            "      DP   DP_ep   DP_aw\n"
        )

        # Print output table if results exist
        body = ''
        if self.bgdata is not None and not self.bgdata.empty:
            display_df = self.bgdata.reset_index().copy()
            body = display_df.to_string(index=False, header=False,
                                        formatters=fmts)

        return header_note + body

    def savefile(self, filepath='bluegreen_output.iwp'):
        """Save BlueGreen output data to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'bluegreen_output.iwp')

        Raises
        ------
        FileNotFoundError
            If the filepath is invalid or cannot be opened.
        """
        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for output data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def plot(self, var='ETa', filename=None):
        """Time series plot of a variable and its EP/AW components.
        saves the figure if filename is provided. Returns plt.show().

        Parameters
        ----------
        var : str
            Variable name in res_pyfao.bgdata 
            (e.g., 'ETa', 'T', 'E', 'Runoff', 'DP').
        filename : str, optional
            If provided, save the figure to this filename (PNG, 300 DPI)
        """

        if self.bgdata is None or self.bgdata.empty:
            raise ValueError("No BlueGreen data available. " \
                             "Please run the model first.")
        ep = f"{var}_ep"
        aw = f"{var}_aw"
        for col in [var, ep, aw]:
            if col not in self.bgdata.columns:
                raise ValueError(f"Column '{col}' not found in result. "
                                 "Try another variable.")

        # Create figure with higher quality
        fig, ax = plt.subplots(figsize=(14, 6), dpi=200)

        # Plot filled areas for components
        if self.bgdata[aw].sum() > self.bgdata[ep].sum():
            z_aw, z_ep = 0, 3  # AW on bottom, EP on top
        else:
            z_ep, z_aw = 0, 3  # EP on bottom, AW on top
        ax.fill_between(self.bgdata['Date'], 0, self.bgdata[ep],
                        color='#2ecc71', alpha=0.4, zorder=z_ep,
                        label=f'{var} (eff. precip.)')
        ax.fill_between(self.bgdata['Date'], 0, self.bgdata[aw],
                        color='#3498db', alpha=0.4, zorder=z_aw,
                        label=f'{var} (applied water)')
        ax.plot(self.bgdata['Date'], self.bgdata[var], linestyle='--',
                lw=2.5, color='black', label=f'{var} (Total)', zorder=5)

        # Secondary y-axis for precipitation and irrigation
        ax01 = ax.twinx()
        ax01.bar(self.bgdata['Date'], self.bgdata['Rain'], width=0.5,
                 color="#05EF6E", alpha=0.4, label='Rainfall',
                 edgecolor='none')
        ax01.bar(self.bgdata['Date'], self.bgdata["Irrig"], width=0.5,
                 color="#0B7BD7", alpha=0.4, label='Irrigation',
                 edgecolor='none')
        ax01.set_ylim(self.bgdata[['Rain', 'Irrig']].max().max()*1.2, 0)
        ax01.set_ylabel("Rainfall & Irrigation (mm)", fontsize=11,
                        fontweight='bold', rotation=270, labelpad=20)
        ax01.tick_params(axis='y', labelsize=10)
        ax01.spines['right'].set_linewidth(1.5)

        # Primary axis styling
        ax.set_xticks(self.bgdata['Date'][::15].to_list())
        ax.set_xlabel("Date", fontsize=12, fontweight='bold')
        ax.set_ylabel(f"{var} (mm/day)", fontsize=12, fontweight='bold')
        ax.grid(axis='y', linestyle='--', alpha=0.3, linewidth=0.8)
        ax.set_ylim(bottom=0)
        ax.set_xlim(self.bgdata['Date'].min(),self.bgdata['Date'].max())

        # Format x-axis dates
        ax.tick_params(axis='x', rotation=30, labelsize=10)
        ax.tick_params(axis='y', labelsize=10)

        # Spines
        for spine in ['top', 'right', 'left', 'bottom']:
            ax.spines[spine].set_linewidth(1.5)

        # Combined legend
        lines1, labels1 = ax.get_legend_handles_labels()
        lines2, labels2 = ax01.get_legend_handles_labels()
        ax.legend(lines1+lines2, labels1 + labels2, loc='upper center',                 bbox_to_anchor=(0.5, 1.15), ncol=5, frameon=True, 
                  fontsize=10, fancybox=True,)

        plt.tight_layout()

        if filename:
            plt.savefig(filename, dpi=300, bbox_inches='tight',
                        facecolor='white')

        return plt.show()



"""
########################################################################
The tables.py module includes the FAO56Tables class, which contains
suggested values for attributes of the Parameters class as defined in
FAO-56, specifically in Tables 11, 12, 17, and 22 of FAO-56.

FAO-56 Table 11 - Lengths of crop development stages
FAO-56 Table 12 - Single crop coefficients (Kc)
FAO-56 Table 17 - Basal crop coefficients (Kcb)
FAO-56 Table 22 - Maximum root depth (Zr) and depletion fraction (p)

The tables.py module contains the following:
    FAO56Tables - A class for specifying parameter values as defined in
                  FAO-56 tables.

12/12/2024 FAOTables class created by Reagan Ames and Tyler Pokoski
01/08/2025 Modification for release by Kelly Thorp
########################################################################
"""

import pandas as pd
import io

class FAO56Tables:
    """A class for specifying parameter values from FAO-56 tables

    Attributes
    ----------
    table11 : Dataframe
        FAO-56 Table 11 data as str
        index - row number as int
        columns - ['Crop','Lini','Ldev','Lmid','Llate','Total',
                   'Plant Date','Region']
            Crop       - Crop description
            Lini       - Length Stage Initial (days)
            Ldev       - Length Stage Development (days)
            Lmid       - Length Stage Mid (days)
            Llate      - Length Stage End (days)
            Total      - Length growing season (days)
            Plant Date - Month of planting
            Region     - Growing location
    table12 : DataFrame
        FAO-56 Table 12 data as str
        index - row number as int
        columns - ['Crop','Kcmini','Kcmmid','Kcmend','hmax']
            Crop  - Crop description
            Kcmini - Kc Initial
            Kcmmid - Kc Mid
            Kcmend - Kc End
            hmax  - Plant Height Maximum (m)
    table17 : DataFrame
        FAO-56 Table 17 data as str
        index - row number as int
        columns - ['Crop','Kcbini','Kcbmid','Kcbend']
            Crop   - Crop description
            Kcbini - Kcb Initial
            Kcbmid - Kcb Mid
            Kcbend - Kcb End
    table22 : DataFrame
        FAO-56 Table 22 data as str
        index - row number as int
        columns - ['Crop','Zrmax1','Zrmax2','pbase']
            Crop   - Crop description
            Zrmax1 - Rooting depth maximum 1 (m)
            Zrmax2 - Rooting depth maximum 2 (m)
            pbase  - Depletion fraction
    synonyms : dict
        keys - common synonyms for crop names and other search terms
        values - list of corresponding  search terms in tables

    Methods
    -------
    search11(phrase,usesyn=True)
        Search FAO-56 Table 11 based on search phrase and synonym list
    search12(phrase,usesyn=True)
        Search FAO-56 Table 12 based on search phrase and synonym list
    search17(phrase,usesyn=True)
        Search FAO-56 Table 17 based on search phrase and synonym list
    search22(phrase,usesyn=True)
        Search FAO-56 Table 22 based on search phrase and synonym list
    """

    def __init__(self):
        """Initialize the FAO56Tables class attributes."""

        #FAO-56 Table 11. Lengths of crop development stages for various planting periods and climatic regions (days)
        #   Crop                         ,Lini,Ldev,Lmid,Llate,Total, Plant Date      , Region
        table11data = """
            broccoli                     ,  35,  45,  40,   15,  135, sep             , california desert usa
            cabbage                      ,  40,  60,  50,   15,  165, sep             , california desert usa
            carrots                      ,  20,  30,  30,   20,  100, oct jan         , arid climate
            carrots                      ,  30,  40,  60,   20,  150, feb mar         , mediterranean
            carrots                      ,  30,  50,  90,   30,  200, oct             , california desert usa
            cauliflower                  ,  35,  50,  40,   15,  140, sep             , california desert usa
            celery                       ,  25,  40,  95,   20,  180, oct             , semiarid
            celery                       ,  25,  40,  45,   15,  125, apr             , mediterranean
            celery                       ,  30,  55, 105,   20,  210, jan             , semiarid
            crucifers                    ,  20,  30,  20,   10,   80, apr             , mediterranean
            crucifers                    ,  25,  35,  25,   10,   95, feb             , mediterranean
            crucifers                    ,  30,  35,  90,   40,  195, oct nov         , mediterranean
            lettuce                      ,  20,  30,  15,   10,   75, apr             , mediterranean
            lettuce                      ,  30,  40,  25,   10,  105, nov jan         , mediterranean
            lettuce                      ,  25,  35,  30,   10,  100, oct nov         , arid region
            lettuce                      ,  35,  50,  45,   10,  140, feb             , mediterranean
            onion dry                    ,  15,  25,  70,   40,  150, apr             , mediterranean
            onion dry                    ,  20,  35, 110,   45,  210, oct jan         , arid region california
            onion green                  ,  25,  30,  10,    5,   70, apr may         , mediterranean
            onion green                  ,  20,  45,  20,   10,   95, oct             , arid region
            onion green                  ,  30,  55,  55,   40,  180, mar             , california usa
            onion seed                   ,  20,  45, 165,   45,  275, sep             , california desert usa
            spinach                      ,  20,  20,  20,    5,   65, apr sep oct     , mediterranean
            spinach                      ,  20,  30,  40,   10,  100, nov             , arid region
            radish                       ,   5,  10,  15,    5,   35, mar apr         , mediterranean europe
            radish                       ,  10,  10,  15,    5,   40, winter          , arid region
            eggplant                     ,  30,  40,  40,   20,  130, oct             , arid region
            eggplant                     ,  30,  45,  40,   25,  140, may jun         , mediterranean
            sweet peppers bell           ,  30,  35,  40,   20,  125, apr jun         , europe mediterranean
            sweet peppers bell           ,  30,  40, 110,   30,  210, oct             , arid region
            tomato                       ,  30,  40,  40,   25,  135, jan             , arid region
            tomato                       ,  35,  40,  50,   30,  155, apr may         , california usa
            tomato                       ,  25,  40,  60,   30,  155, jan             , california desert usa
            tomato                       ,  35,  45,  70,   30,  180, oct nov         , arid region
            tomato                       ,  30,  40,  45,   30,  145, apr may         , mediterranean
            cantaloupe                   ,  30,  45,  35,   10,  120, jan             , california usa
            cantaloupe                   ,  10,  60,  25,   25,  120, aug             , california usa
            cucumber                     ,  20,  30,  40,   15,  105, june aug        , arid region
            cucumber                     ,  25,  35,  50,   20,  130, nov feb         , arid region
            pumpkin winter squash        ,  20,  30,  30,   20,  100, mar aug         , mediterranean
            pumpkin winter squash        ,  25,  35,  35,   25,  120, jun             , europe
            squash zucchini              ,  25,  35,  25,   15,  100, apr dec         , mediterranean arid region
            squash zucchini              ,  20,  30,  25,   15,   90, may jun         , mediterranean europe
            sweet melons                 ,  25,  35,  40,   20,  120, may             , mediterranean
            sweet melons                 ,  30,  30,  50,   30,  140, mar             , california usa
            sweet melons                 ,  15,  40,  65,   15,  135, aug             , california desert usa
            sweet melons                 ,  30,  45,  65,   20,  160, dec jan         , arid region
            water melons                 ,  20,  30,  30,   30,  110, apr             , italy
            water melons                 ,  10,  20,  20,   30,   80, may aug         , near east desert
            beets table                  ,  15,  25,  20,   10,   70, apr may         , mediterranean
            beets table                  ,  25,  30,  25,   10,   90, feb mar         , mediterranean arid
            cassava year 1               ,  20,  40,  90,   60,  210, rainy season    , tropical regions
            cassava year 2               , 150,  40, 110,   60,  360, rainy season    , tropical regions
            potato                       ,  25,  30,  40,   30,  125, jan nov         , semiarid climate
            potato                       ,  25,  30,  45,   30,  130, may             , continental climate
            potato                       ,  30,  35,  50,   30,  145, apr             , europe
            potato                       ,  45,  30,  70,   20,  165, apr may         , idaho usa
            potato                       ,  30,  35,  50,   25,  140, dec             , california desert usa
            sweet potato                 ,  20,  30,  60,   40,  150, apr             , mediterranean
            sweet potato                 ,  15,  30,  50,   30,  125, rainy season    , tropical regions
            sugarbeet                    ,  30,  45,  90,   15,  180, mar             , california usa
            sugarbeet                    ,  25,  30,  90,   10,  155, jun             , california usa
            sugarbeet                    ,  25,  65, 100,   65,  255, sep             , california desert usa
            sugarbeet                    ,  50,  40,  50,   40,  180, apr             , idaho usa
            sugarbeet                    ,  25,  35,  50,   50,  160, may             , mediterranean
            sugarbeet                    ,  45,  75,  80,   30,  230, nov             , mediterranean
            sugarbeet                    ,  35,  60,  70,   40,  205, nov             , arid regions
            beans green                  ,  20,  30,  30,   10,   90, feb mar         , california mediterranean
            beans green                  ,  15,  25,  25,   10,   75, aug sep         , california egypt lebanon
            beans dry                    ,  20,  30,  40,   20,  110, may jun         , continental climates
            beans dry                    ,  15,  25,  35,   20,   95, jun             , pakistan california
            beans dry                    ,  25,  25,  30,   20,  100, jun             , idaho usa
            faba bean broad              ,  15,  25,  35,   15,   90, may             , europe
            faba bean broad              ,  20,  30,  35,   15,  100, mar apr         , mediterranean
            faba bean broad dry          ,  90,  45,  40,   60,  235, nov             , europe
            faba bean broad green        ,  90,  45,  40,    0,  175, nov             , europe
            green gram cowpeas           ,  20,  30,  30,   20,  110, mar             , mediterranean
            groundnut                    ,  25,  35,  45,   25,  130, dry season      , west africa
            groundnut                    ,  35,  35,  35,   35,  140, may             , high latitudes
            groundnut                    ,  35,  45,  35,   25,  140, may jun         , mediterranean
            lentil                       ,  20,  30,  60,   40,  150, apr             , europe
            lentil                       ,  25,  35,  70,   40,  170, oct nov         , arid region
            peas                         ,  15,  25,  35,   15,   90, may             , europe
            peas                         ,  20,  30,  35,   15,  100, mar apr         , mediterranean
            peas                         ,  35,  25,  30,   20,  110, apr             , idaho usa
            soybeans                     ,  15,  15,  40,   15,   85, dec             , tropics
            soybeans                     ,  20,  35,  60,   25,  140, may             , central usa
            soybeans                     ,  20,  25,  75,   30,  150, jun             , japan
            artichoke                    ,  40,  40, 250,   30,  360, apr first year  , california
            artichoke                    ,  20,  25, 250,   30,  325, may second year , california
            asparagus                    ,  50,  30, 100,   50,  230, feb             , warm winter
            asparagus                    ,  90,  30, 200,   45,  365, feb             , mediterranean
            cotton                       ,  30,  50,  60,   55,  195, mar may         , egypt pakistan california
            cotton                       ,  45,  90,  45,   45,  225, mar             , california desert usa
            cotton                       ,  30,  50,  60,   55,  195, sep             , yemen
            cotton                       ,  30,  50,  55,   45,  180, apr             , texas
            flax                         ,  25,  35,  50,   40,  150, apr             , europe
            flax                         ,  30,  40, 100,   50,  220, oct             , arizona
            castor beans                 ,  25,  40,  65,   50,  180, mar             , semiarid climates
            castor beans                 ,  20,  40,  50,   25,  135, nov             , indonesia
            safflower                    ,  20,  35,  45,   25,  125, apr             , california usa
            safflower                    ,  25,  35,  55,   30,  145, mar             , high latitudes
            safflower                    ,  35,  55,  60,   40,  190, oct nov         , arid region
            sesame                       ,  20,  30,  40,   20,  100, jun             , china
            sunflower                    ,  25,  35,  45,   25,  130, apr may         , mediterranean california
            barley oats wheat            ,  15,  25,  50,   30,  120, nov             , central india
            barley oats wheat            ,  20,  25,  60,   30,  135, mar apr         , 35-45 degrees latitude
            barley oats wheat            ,  15,  30,  65,   40,  150, jul             , east africa
            barley oats wheat            ,  40,  30,  40,   20,  130, apr             , east africa
            barley oats wheat            ,  40,  60,  60,   40,  200, nov             , east africa
            barley oats wheat            ,  20,  50,  60,   30,  160, dec             , california desert usa
            winter wheat                 ,  20,  60,  70,   30,  180, dec             , california usa
            winter wheat                 ,  30, 140,  40,   30,  240, nov             , mediterranean
            winter wheat                 , 160,  75,  75,   25,  335, oct             , idaho usa
            grains small                 ,  20,  30,  60,   40,  150, apr             , mediterranean
            grains small                 ,  25,  35,  65,   40,  165, oct nov         , pakistan arid region
            maize grain                  ,  30,  50,  60,   40,  180, apr             , east africa
            maize grain                  ,  25,  40,  45,   30,  140, dec jan         , arid climate
            maize grain                  ,  20,  35,  40,   30,  125, jun             , nigeria humid
            maize grain                  ,  20,  35,  40,   30,  125, oct             , india dry cool
            maize grain                  ,  30,  40,  50,   30,  150, apr             , spain spring summer california
            maize grain                  ,  30,  40,  50,   50,  170, apr             , idaho usa
            maize sweet                  ,  20,  20,  30,   10,   80, mar             , philippines
            maize sweet                  ,  20,  25,  25,   10,   80, may jun         , mediterranean
            maize sweet                  ,  20,  30,  30,   10,   90, oct dec         , arid climate
            maize sweet                  ,  30,  30,  30,   10,  110, apr             , idaho usa
            maize sweet                  ,  20,  40,  70,   10,  140, jan             , california desert usa
            millet                       ,  15,  25,  40,   25,  105, jun             , pakistan
            millet                       ,  20,  30,  55,   35,  140, apr             , central usa
            sorghum                      ,  20,  35,  40,   30,  130, may june        , usa pakistan mediterranean
            sorghum                      ,  20,  35,  45,   30,  140, mar apr         , arid region
            rice                         ,  30,  30,  60,   30,  150, dec may         , tropics mediterranean
            rice                         ,  30,  30,  80,   40,  180, may             , tropics
            alfalfa total season         ,  10,  30,    ,     ,     ,                 , last -4oC in spring until first -4oC in fall
            alfalfa first cutting cycle  ,  10,  20,  20,   10,   60, jan             , california usa
            alfalfa first cutting cycle  ,  10,  30,  25,   10,   75, apr last -4oC   , idaho usa
            alfalfa other cutting cycles ,   5,  10,  10,    5,   30, mar             , california usa
            alfalfa other cutting cycles ,   5,  20,  10,   10,   45, jun             , idaho usa
            bermuda for seed             ,  10,  25,  35,   35,  105, mar             , california desert usa
            bermuda hay several cuttings ,  10,  15,  75,   35,  135,                 , california desert usa
            grass pasture                ,  10,  20,    ,     ,     ,                 , 7 days before last -4oC in spring until 7 days after first -4oC in fall
            sudan first cutting cycle    ,  25,  25,  15,   10,   75, apr             , california desert usa
            sudan other cutting cycles   ,   3,  15,  12,    7,   37, jun             , california desert usa
            sugarcane virgin             ,  35,  60, 190,  120,  405,                 , low latitudes
            sugarcane virgin             ,  50,  70, 220,  140,  480,                 , tropics
            sugarcane virgin             ,  75, 105, 330,  210,  720,                 , hawaii usa
            sugarcane ratoon             ,  25,  70, 135,   50,  280,                 , low latitudes
            sugarcane ratoon             ,  30,  50, 180,   60,  320,                 , tropics
            sugarcane ratoon             ,  35, 105, 210,   70,  420,                 , hawaii usa
            banana first year            , 120,  90, 120,   60,  390, mar             , mediterranean
            banana second year           , 120,  60, 180,    5,  365, feb             , mediterranean
            pineapple                    ,  60, 120, 600,   10,  790,                 , hawaii usa
            grapes                       ,  20,  40, 120,   60,  240, apr             , low latitudes
            grapes                       ,  20,  50,  75,   60,  205, mar             , california usa
            grapes                       ,  20,  50,  90,   20,  180, may             , high latitudes
            grapes                       ,  30,  60,  40,   80,  210, apr             , mid latitudes wine
            hops                         ,  25,  40,  80,   10,  155, apr             , idaho usa
            citrus                       ,  60,  90, 120,   95,  365, jan             , mediterranean
            deciduous orchard            ,  20,  70,  90,   30,  210, mar             , high latitudes
            deciduous orchard            ,  20,  70, 120,   60,  270, mar             , low latitudes
            deciduous orchard            ,  30,  50, 130,   30,  240, mar             , california usa
            olives                       ,  30,  90,  60,   90,  270, mar             , mediterranean
            pistachios                   ,  20,  60,  30,   40,  150, feb             , mediterranean
            walnuts                      ,  20,  10, 130,   30,  190, apr             , utah usa
            wetlands cattails bulrush    ,  10,  30,  80,   20,  140, may             , utah usa killing frost
            wetlands cattails bulrush    , 180,  60,  90,   35,  365, nov             , florida usa
            wetlands short vegetation    , 180,  60,  90,   35,  365, nov             , frost-free climate"""
        self.table11 = pd.read_csv(io.StringIO(table11data),
                                   names=['Crop','Lini','Ldev','Lmid','Llate','Total','Plant Date','Region'],
                                   dtype=str)
        for col in self.table11.columns:
            self.table11[col] = self.table11[col].str.strip()

        #FAO-56 Table 12. Single (time-averaged) crop coefficients, Kcm, and mean maximum plant
        #heights for non stressed, well-managed crops in subhumid climates (RHmin=45%, u2=2m/s)
        #for use with the FAO Penman-Monteith ETo.
        #   Crop                                                          ,Kcini,Kcmid,Kcend,hmax
        table12data = """
            small vegetables                                              , 0.7 , 1.05, 0.95, 
            broccoli                                                      , 0.7 , 1.05, 0.95, 0.3
            brussel sprouts                                               , 0.7 , 1.05, 0.95, 0.4
            cabbage                                                       , 0.7 , 1.05, 0.95, 0.4
            carrots                                                       , 0.7 , 1.05, 0.95, 0.3
            cauliflower                                                   , 0.7 , 1.05, 0.95, 0.4
            celery                                                        , 0.7 , 1.05, 1.00, 0.6
            garlic                                                        , 0.7 , 1.00, 0.70, 0.3
            lettuce                                                       , 0.7 , 1.00, 0.95, 0.3
            onions dry                                                    , 0.7 , 1.05, 0.75, 0.4
            onions green                                                  , 0.7 , 1.00, 1.00, 0.3
            onions seed                                                   , 0.7 , 1.05, 0.80, 0.5
            spinach                                                       , 0.7 , 1.00, 0.95, 0.3
            radish                                                        , 0.7 , 0.90, 0.85, 0.3
            vegetables solanum family solanaceae                          , 0.6 , 1.15, 0.80, 
            eggplant                                                      , 0.6 , 1.05, 0.90, 0.8
            sweet peppers bell                                            , 0.6 , 1.05, 0.90, 0.7
            tomato                                                        , 0.6 , 1.15, 0.80, 0.6
            vegetables cucumber family cucurbitaceae                      , 0.5 , 1.00, 0.80, 
            cantaloupe                                                    , 0.5 , 0.85, 0.60, 0.3
            cucumber fresh market                                         , 0.6 , 1.00, 0.75, 0.3
            cucumber machine harvest                                      , 0.5 , 1.00, 0.90, 0.3
            pumpkin winter squash                                         , 0.5 , 1.00, 0.80, 0.4
            squash zucchini                                               , 0.5 , 0.95, 0.75, 0.3
            sweet melons                                                  , 0.5 , 1.05, 0.75, 0.4
            watermelon                                                    , 0.4 , 1.00, 0.75, 0.4
            roots tubers                                                  , 0.5 , 1.10, 0.95, 
            beets table                                                   , 0.5 , 1.05, 0.95, 0.4
            cassava year 1                                                , 0.3 , 0.80, 0.30, 1.0
            cassava year 2                                                , 0.3 , 1.10, 0.50, 1.5
            parsnip                                                       , 0.5 , 1.05, 0.95, 0.4
            potato                                                        , 0.5 , 1.15, 0.75, 0.6
            sweet potato                                                  , 0.5 , 1.15, 0.65, 0.4
            turnip rutabaga                                               , 0.5 , 1.10, 0.95, 0.6
            sugar beet                                                    , 0.35, 1.20, 0.70, 0.5
            legumes leguminosae                                           , 0.4 , 1.15, 0.55, 
            beans green                                                   , 0.5 , 1.05, 0.90, 0.4
            beans dry pulses                                              , 0.4 , 1.15, 0.35, 0.4
            chickpea                                                      , 0.4 , 1.00, 0.35, 0.4
            faba bean broad fresh                                         , 0.5 , 1.15, 1.10, 0.8
            faba bean broad dry seed                                      , 0.5 , 1.15, 0.30, 0.8
            garbanzo                                                      , 0.4 , 1.15, 0.35, 0.8
            green gram cowpeas harvested fresh                            , 0.4 , 1.05, 0.60, 0.4
            green gram cowpeas harvested dry                              , 0.4 , 1.05, 0.35, 0.4
            groundnut peanut                                              , 0.4 , 1.15, 0.60, 0.4
            lentil                                                        , 0.4 , 1.10, 0.30, 0.5
            peas fresh                                                    , 0.5 , 1.15, 1.10, 0.5
            peas dry seed                                                 , 0.5 , 1.15, 0.30, 0.5
            soybeans                                                      , 0.5 , 1.15, 0.50, 0.75
            perennial vegetables winter dormancy bare mulched soil        , 0.5 , 1.00, 0.80, 
            artichokes                                                    , 0.5 , 1.00, 0.95, 0.7
            asparagus                                                     , 0.5 , 0.95, 0.30, 0.5
            mint                                                          , 0.60, 1.15, 1.10, 0.7
            strawberries                                                  , 0.40, 0.85, 0.75, 0.2
            fiber crops                                                   , 0.35,     ,     , 
            cotton                                                        , 0.35, 1.17, 0.60, 1.35
            flax                                                          , 0.35, 1.10, 0.25, 1.2
            sisal                                                         , 0.35, 0.55, 0.55, 1.5
            oil crops                                                     , 0.35, 1.15, 0.35, 
            castorbean ricinus                                            , 0.35, 1.15, 0.55, 0.3
            rapeseed canola                                               , 0.35, 1.10, 0.35, 0.6
            safflower                                                     , 0.35, 1.10, 0.25, 0.8
            sesame                                                        , 0.35, 1.10, 0.25, 1.0
            sunflower                                                     , 0.35, 1.10, 0.35, 2.0
            cereals                                                       , 0.3 , 1.15, 0.4 , 
            barley                                                        , 0.3 , 1.15, 0.25, 1
            oats                                                          , 0.3 , 1.15, 0.25, 1
            spring wheat                                                  , 0.3 , 1.15, 0.25, 1
            winter wheat frozen soils                                     , 0.4 , 1.15, 0.25, 1
            winter wheat non-frozen soils                                 , 0.7 , 1.15, 0.25, 1
            maize field grain corn 18% moisture                           , 0.3 , 1.20, 0.35, 2
            maize field grain corn harvested wet                          , 0.3 , 1.20, 0.60, 2
            maize sweet corn fresh                                        , 0.3 , 1.15, 1.05, 1.5
            maize sweet corn dry                                          , 0.3 , 1.15, 0.35, 1.5
            millet                                                        , 0.3 , 1.00, 0.30, 1.5
            sorghum grain                                                 , 0.3 , 1.05, 0.55, 1.5
            sorgum sweet                                                  , 0.3 , 1.20, 1.05, 3
            rice                                                          , 1.05, 1.20, 0.75, 1
            alfalfa hay averaged cutting effects                          , 0.40, 0.95, 0.90, 0.7
            alfalfa hay individual cutting periods                        , 0.40, 1.20, 1.15, 0.7
            alfalfa for seed                                              , 0.40, 0.50, 0.50, 0.7
            bermuda hay averaged cutting effects                          , 0.55, 1.00, 0.85, 0.35
            bermuda hay spring crop for seed                              , 0.35, 0.90, 0.65, 0.4
            clover hay berseem averaged cutting effects                   , 0.40, 0.90, 0.85, 0.6
            clover hay individual cutting periods                         , 0.40, 1.15, 1.10, 0.6
            rye grass hay averaged cutting effects                        , 0.95, 1.05, 1.00, 0.3
            sudangrass hay annual averaged cutting effects                , 0.50, 0.90, 0.85, 1.2
            sudangrass individual cutting periods                         , 0.50, 1.15, 1.10, 1.2
            grazing pasture rotated                                       , 0.40, 0.90, 0.85, 0.22
            grazing pasture extensive                                     , 0.30, 0.75, 0.75, 0.10
            turfgrass cool season                                         , 0.90, 0.95, 0.95, 0.10
            turfgrass warm season                                         , 0.80, 0.85, 0.85, 0.10
            sugarcane                                                     , 0.40, 1.25, 0.75, 3
            banana first year                                             , 0.50, 1.10, 1.00, 3
            banana second year                                            , 1.00, 1.20, 1.10, 4
            cacao                                                         , 1.00, 1.05, 1.05, 3
            coffee bare ground cover                                      , 0.90, 0.95, 0.95, 2.5
            coffee with weeds                                             , 1.05, 1.10, 1.10, 2.5
            date palms                                                    , 0.90, 0.95, 0.95, 8
            palm trees                                                    , 0.95, 1.00, 1.00, 8
            pineapple bare soil                                           , 0.50, 0.30, 0.30, 0.9
            pineapple with grass cover                                    , 0.50, 0.50, 0.50, 0.9
            rubber trees                                                  , 0.95, 1.00, 1.00, 10
            tea non-shaded                                                , 0.95, 1.00, 1.00, 1.5
            tea shaded                                                    , 1.10, 1.15, 1.15, 2
            berries bushes                                                , 0.30, 1.05, 0.50, 1.5
            grapes table raisin                                           , 0.30, 0.85, 0.45, 2
            grapes wine                                                   , 0.30, 0.70, 0.45, 1.75
            hops                                                          , 0.3 , 1.05, 0.85, 5
            almonds no ground cover                                       , 0.40, 0.90, 0.65, 5
            apples cherries pears no ground cover killing frost           , 0.45, 0.95, 0.70, 4
            apples cherries pears no ground cover no frosts               , 0.60, 0.95, 0.75, 4
            apples cherries pears active ground cover killing frost       , 0.50, 1.20, 0.95, 4
            apples cherries pears active ground cover no frosts           , 0.80, 1.20, 0.85, 4
            apricots peaches stone fruit no ground cover killing frost    , 0.45, 0.90, 0.65, 3
            apricots peaches stone fruit no ground cover no frosts        , 0.55, 0.90, 0.65, 3
            apricots peaches stone fruit active ground cover killing frost, 0.50, 1.15, 0.90, 3
            apricots peaches stone fruit active ground cover no frosts    , 0.80, 1.15, 0.85, 3
            avocado no ground cover                                       , 0.60, 0.85, 0.75, 3
            citrus no ground cover 70% canopy                             , 0.70, 0.65, 0.70, 4
            citrus no ground cover 50% canopy                             , 0.65, 0.60, 0.65, 3
            citrus no ground cover 20% canopy                             , 0.50, 0.45, 0.55, 2
            citrus active ground cover or weeds 70% canopy                , 0.75, 0.70, 0.75, 4
            citrus active ground cover or weeds 50% canopy                , 0.80, 0.80, 0.80, 3
            citrus active ground cover or weeds 20% canopy                , 0.85, 0.85, 0.85, 2
            conifer trees                                                 , 1.00, 1.00, 1.00, 10
            kiwi                                                          , 0.40, 1.05, 1.05, 3
            olives 40% to 60% ground coverage by canopy                   , 0.65, 0.70, 0.70, 4
            pistachios no ground cover                                    , 0.40, 1.10, 0.45, 4
            walnut orchard                                                , 0.50, 1.10, 0.65, 4.5
            cattails bulrushes killing frost                              , 0.30, 1.20, 0.30, 2
            cattails bulrushes no frost                                   , 0.60, 1.20, 0.60, 2
            short vegetation no frost                                     , 1.05, 1.10, 1.10, 0.3
            reed swamp standing water                                     , 1.00, 1.20, 1.00, 2
            reed swamp moist soil                                         , 0.90, 1.20, 0.70, 2
            open water < 2 m depth or in subhumid climates or tropics     ,     , 1.05, 1.05, 
            open water > 5 m depth clear of turbidity temperate climate   ,     , 0.65, 1.25,  """
        self.table12 = pd.read_csv(io.StringIO(table12data),
                                   names=['Crop','Kcmini','Kcmmid','Kcmend','hmax'],
                                   dtype=str)
        for col in self.table12.columns:
            self.table12[col] = self.table12[col].str.strip()

        #FAO-56 Table 17. Basal crop coefficients, Kcb, for non stressed, well-managed crops in
        #subhumid climates (RHmin = 45%, u2 = 2 m/s) for use with the FAO Penman-Montieth ETo
        #   Crop                                                          Kcbini,Kcbmid,Kcbend
        table17data = """
            small vegetables                                              , 0.15,  0.95,  0.85
            broccoli                                                      , 0.15,  0.95,  0.85
            brussel sprouts                                               , 0.15,  0.95,  0.85
            cabbage                                                       , 0.15,  0.95,  0.85
            carrots                                                       , 0.15,  0.95,  0.85
            cauliflower                                                   , 0.15,  0.95,  0.85
            celery                                                        , 0.15,  0.95,  0.90
            garlic                                                        , 0.15,  0.90,  0.60
            lettuce                                                       , 0.15,  0.90,  0.90
            onions dry                                                    , 0.15,  0.95,  0.65
            onions green                                                  , 0.15,  0.90,  0.90
            onions seed                                                   , 0.15,  1.05,  0.70
            spinach                                                       , 0.15,  0.90,  0.85
            radishes                                                      , 0.15,  0.85,  0.75
            vegetables solanum family solanaceae                          , 0.15,  1.10,  0.70
            eggplant                                                      , 0.15,  1.00,  0.80
            sweet peppers bell                                            , 0.15,  1.00,  0.80
            tomato                                                        , 0.15,  1.10,  0.70
            vegetables cucumber family cucurbitaceae                      , 0.15,  0.95,  0.70
            cantaloupe                                                    , 0.15,  0.75,  0.50
            cucumber fresh market                                         , 0.15,  0.95,  0.70
            cucumber machine harvest                                      , 0.15,  0.95,  0.80
            pumpkin winter squash                                         , 0.15,  0.95,  0.70
            squash zucchini                                               , 0.15,  0.90,  0.70
            sweet melons                                                  , 0.15,  1.00,  0.70
            watermelon                                                    , 0.15,  0.95,  0.70
            roots tubers                                                  , 0.15,  1.00,  0.85
            beets table                                                   , 0.15,  0.95,  0.85
            cassava year 1                                                , 0.15,  0.70,  0.20
            cassava year 2                                                , 0.15,  1.00,  0.45
            parsnip                                                       , 0.15,  0.95,  0.85
            potato                                                        , 0.15,  1.10,  0.65
            sweet potato                                                  , 0.15,  1.10,  0.55
            turnip rutabaga                                               , 0.15,  1.00,  0.85
            sugar beet                                                    , 0.15,  1.15,  0.50
            legumes leguminosae                                           , 0.15,  1.10,  0.50
            beans green                                                   , 0.15,  1.00,  0.80
            beans dry pulses                                              , 0.15,  1.10,  0.25
            chickpea                                                      , 0.15,  0.95,  0.25
            faba bean broad fresh                                         , 0.15,  1.10,  1.05
            faba bean dry seed                                            , 0.15,  1.10,  0.20
            garbanzo                                                      , 0.15,  1.05,  0.25
            green gram cowpeas fresh                                      , 0.15,  1.00,  0.55
            green gram cowpeas dry                                        , 0.15,  1.00,  0.25
            groundnut peanut                                              , 0.15,  1.10,  0.50
            lentil                                                        , 0.15,  1.05,  0.20
            peas fresh                                                    , 0.15,  1.10,  1.05
            peas dry seed                                                 , 0.15,  1.10,  0.20
            soybeans                                                      , 0.15,  1.10,  0.30
            artichokes                                                    , 0.15,  0.95,  0.90
            asparagus                                                     , 0.15,  0.90,  0.20
            mint                                                          , 0.40,  1.10,  1.05
            strawberries                                                  , 0.30,  0.80,  0.70
            fiber crops                                                   , 0.15,      , 
            cotton                                                        , 0.15,  1.12,  0.45
            flax                                                          , 0.15,  1.05,  0.20
            sisal                                                         , 0.15,  0.55,  0.55
            oil crops                                                     , 0.15,  1.10,  0.25
            castorbean ricinus                                            , 0.15,  1.10,  0.45
            rapeseed canola                                               , 0.15,  1.10,  0.25
            safflower                                                     , 0.15,  1.10,  0.20
            sesame                                                        , 0.15,  1.05,  0.20
            sunflower                                                     , 0.15,  1.10,  0.25
            cereals                                                       , 0.15,  1.10,  0.25
            barley                                                        , 0.15,  1.10,  0.15
            oats                                                          , 0.15,  1.10,  0.15
            spring wheat                                                  , 0.15,  1.10,  0.15
            winter wheat                                                  , 0.15,  1.10,  0.15
            maize field grain corn 18% moisture                           , 0.15,  1.15,  0.15
            maize field grain corn harvested wet                          , 0.15,  1.15,  0.50
            maize sweet corn fresh                                        , 0.15,  1.10,  1.00
            maize sweet corn dry                                          , 0.15,  1.10,  0.15
            millet                                                        , 0.15,  0.95,  0.20
            sorghum grain                                                 , 0.15,  1.00,  0.35
            sorghum sweet                                                 , 0.15,  1.15,  1.00
            rice                                                          , 1.00,  1.15,  0.57
            alfalfa hay individual cutting periods                        , 0.30,  1.15,  1.10
            alfalfa hay for seed                                          , 0.30,  0.45,  0.45
            bermuda hay averaged cutting effects                          , 0.50,  0.95,  0.80
            bermuda hay spring crop for seed                              , 0.15,  0.85,  0.60
            clover hay berseem individual cutting periods                 , 0.30,  1.10,  1.05
            rye grass hay averaged cutting effects                        , 0.85,  1.00,  0.95
            sudangrass hay annual individual cutting periods              , 0.30,  1.10,  1.05
            grazing pasture rotated grazing                               , 0.30,  0.90,  0.80
            grazing pasture extensive grazing                             , 0.30,  0.70,  0.70
            turfgrass cool season                                         , 0.85,  0.90,  0.90
            turfgrass warm season                                         , 0.75,  0.80,  0.80
            sugarcane                                                     , 0.15,  1.20,  0.70
            banana first year                                             , 0.15,  1.05,  0.90
            banana second year                                            , 0.60,  1.10,  1.05
            cacao                                                         , 0.90,  1.00,  1.00
            coffee bare ground cover                                      , 0.80,  0.90,  0.90
            coffee with weeds                                             , 1.00,  1.05,  1.05
            date palms                                                    , 0.80,  0.85,  0.85
            palm trees                                                    , 0.85,  0.90,  0.90
            pineapple multiyear crop bare soil                            , 0.15,  0.25,  0.25
            pineapple multiyear crop with grass cover                     , 0.30,  0.45,  0.45
            rubber trees                                                  , 0.85,  0.90,  0.90
            tea nonshaded                                                 , 0.90,  0.95,  0.90
            tea shaded                                                    , 1.00,  1.10,  1.05
            berries bushes                                                , 0.20,  1.00,  0.40
            grapes table raisin                                           , 0.15,  0.80,  0.40
            grapes wine                                                   , 0.15,  0.65,  0.40
            hops                                                          , 0.15,  1.00,  0.80
            almonds no ground cover                                       , 0.20,  0.85,  0.60
            apples cherries pears no ground cover killing frost           , 0.35,  0.90,  0.65
            apples cherries pears no ground cover no frosts               , 0.50,  0.90,  0.70
            apples cherries pears active ground cover killing frost       , 0.45,  1.15,  0.90
            apples cherries pears active ground cover no frosts           , 0.75,  1.15,  0.80
            apricots peaches stone fruit no ground cover killing frost    , 0.35,  0.85,  0.60
            apricots peaches stone fruit no ground cover no frosts        , 0.45,  0.85,  0.60
            apricots peaches stone fruit active ground cover killing frost, 0.45,  1.10,  0.85
            apricots peaches stone fruit active ground cover no frosts    , 0.75,  1.10,  0.80
            avocado no ground cover                                       , 0.50,  0.80,  0.70
            citrus no ground cover 70% canopy                             , 0.65,  0.60,  0.65
            citrus no ground cover 50% canopy                             , 0.60,  0.55,  0.60
            citrus no ground cover 20% canopy                             , 0.45,  0.40,  0.50
            citrus with active ground cover or weeds 70% canopy           , 0.75,  0.70,  0.75
            citrus with active ground cover or weeds 50% canopy           , 0.75,  0.75,  0.75
            citrus with active ground cover or weeds 20% canopy           , 0.80,  0.80,  0.85
            conifer trees                                                 , 0.95,  0.95,  0.95
            kiwi                                                          , 0.20,  1.00,  1.00
            olives 40% to 60% ground coverage by canopy                   , 0.55,  0.65,  0.65
            pistachios no ground cover                                    , 0.20,  1.05,  0.40
            walnut orchard                                                , 0.40,  1.05,  0.60"""
        self.table17 = pd.read_csv(io.StringIO(table17data),
                                   names=['Crop','Kcbini','Kcbmid','Kcbend'],
                                   dtype=str)
        for col in self.table17.columns:
            self.table17[col] = self.table17[col].str.strip()

        #FAO-56 Table 22. Ranges of maximum effective rooting depth (Zr)
        #and soil water depletion fraction for no stress (p) for common
        #crops
        #   Crop                                      Zrmax1,Zrmax2,pbase
        table22data = """
            broccoli                                   , 0.4,   0.6, 0.45
            brussel sprouts                            , 0.4,   0.6, 0.45
            cabbage                                    , 0.5,   0.8, 0.45
            carrots                                    , 0.5,   1.0, 0.35
            cauliflower                                , 0.4,   0.7, 0.45
            celery                                     , 0.3,   0.5, 0.20
            garlic                                     , 0.3,   0.5, 0.30
            lettuce                                    , 0.3,   0.5, 0.30
            onions dry                                 , 0.3,   0.6, 0.30
            onoins green                               , 0.3,   0.6, 0.30
            onions seed                                , 0.3,   0.6, 0.35
            spinach                                    , 0.3,   0.5, 0.20
            radishes                                   , 0.3,   0.5, 0.30
            eggplant                                   , 0.7,   1.2, 0.45
            sweet peppers bell                         , 0.5,   1.0, 0.30
            tomato                                     , 0.7,   1.5, 0.40
            cantaloupe                                 , 0.9,   1.5, 0.45
            cucumber fresh market                      , 0.7,   1.2, 0.50
            cucumber machine harvest                   , 0.7,   1.2, 0.50
            pumpkin winter squash                      , 1.0,   1.5, 0.35
            squash zucchini                            , 0.6,   1.0, 0.50
            sweet melons                               , 0.8,   1.5, 0.40
            watermelon                                 , 0.8,   1.5, 0.40
            beets table                                , 0.6,   1.0, 0.50
            cassava year 1                             , 0.5,   0.8, 0.35
            cassava year 2                             , 0.7,   1.0, 0.40
            parsnip                                    , 0.5,   1.0, 0.40
            potato                                     , 0.4,   0.6, 0.35
            sweet potato                               , 1.0,   1.5, 0.65
            turnip rutabaga                            , 0.5,   1.0, 0.50
            sugar beet                                 , 0.7,   1.2, 0.55
            beans green                                , 0.5,   0.7, 0.45
            beans dry pulses                           , 0.6,   0.9, 0.45
            beans lima large vines                     , 0.8,   1.2, 0.45
            chickpea                                   , 0.6,   1.0, 0.50
            faba bean broad fresh                      , 0.5,   0.7, 0.45
            faba bean broad dry seed                   , 0.5,   0.7, 0.45
            garbanzo                                   , 0.6,   1.0, 0.45
            green gram cowpeas                         , 0.6,   1.0, 0.45
            groundnut peanut                           , 0.5,   1.0, 0.50
            lentil                                     , 0.6,   0.8, 0.50
            peas fresh                                 , 0.6,   1.0, 0.35
            peas dry seed                              , 0.6,   1.0, 0.40
            soybeans                                   , 0.6,   1.3, 0.50
            artichokes                                 , 0.6,   0.9, 0.45
            asparagus                                  , 1.2,   1.8, 0.45
            mint                                       , 0.4,   0.8, 0.40
            strawberries                               , 0.2,   0.3, 0.20
            cotton                                     , 1.0,   1.7, 0.65
            flax                                       , 1.0,   1.5, 0.50
            sisal                                      , 0.5,   1.0, 0.80
            castorbean ricinus                         , 1.0,   2.0, 0.50
            rapeseed canola                            , 1.0,   1.5, 0.60
            safflower                                  , 1.0,   2.0, 0.60
            sesame                                     , 1.0,   1.5, 0.60
            sunflower                                  , 0.8,   1.5, 0.45
            barley                                     , 1.0,   1.5, 0.55
            oats                                       , 1.0,   1.5, 0.55
            spring wheat                               , 1.0,   1.5, 0.55
            winter wheat                               , 1.5,   1.8, 0.55
            maize field grain corn                     , 1.0,   1.7, 0.55
            maize sweet corn                           , 0.8,   1.2, 0.50
            millet                                     , 1.0,   2.0, 0.55
            sorghum grain                              , 1.0,   2.0, 0.55
            sorghum sweet                              , 1.0,   2.0, 0.50
            rice                                       , 0.5,   1.0, 0.20
            alfalfa for hay                            , 1.0,   2.0, 0.55
            alfalfa for seed                           , 1.0,   3.0, 0.60
            bermuda for hay                            , 1.0,   1.5, 0.55
            spring crop for seed                       , 1.0,   1.5, 0.60
            clover hay berseem                         , 0.6,   0.9, 0.50
            rye grass hay                              , 0.6,   1.0, 0.60
            sudangrass hay annual                      , 1.0,   1.5, 0.55
            grazing pasture rotated                    , 0.5,   1.5, 0.60
            grazing pasture extensive                  , 0.5,   1.5, 0.60
            turfgrass cool season                      , 0.5,   1.0, 0.40
            turfgrass warm season                      , 0.5,   1.0, 0.50
            sugarcane                                  , 1.2,   2.0, 0.65
            banana first year                          , 0.5,   0.9, 0.35
            banana second year                         , 0.5,   0.9, 0.35
            cacao                                      , 0.7,   1.0, 0.30
            coffee                                     , 0.9,   1.5, 0.40
            date palms                                 , 1.5,   2.5, 0.50
            palm trees                                 , 0.7,   1.1, 0.65
            pineapple                                  , 0.3,   0.6, 0.50
            rubber trees                               , 1.0,   1.5, 0.40
            tea non-shaded                             , 0.9,   1.5, 0.40
            tea shaded                                 , 0.9,   1.5, 0.45
            berries bushes                             , 0.6,   1.2, 0.50
            grapes table raisin                        , 1.0,   2.0, 0.35
            grapes wine                                , 1.0,   2.0, 0.45
            hops                                       , 1.0,   1.2, 0.50
            almonds                                    , 1.0,   2.0, 0.40
            apples cherries pears                      , 1.0,   2.0, 0.50
            apricots peaches stone fruit               , 1.0,   2.0, 0.50
            avocado                                    , 0.5,   1.0, 0.70
            citrus 70% canopy                          , 1.2,   1.5, 0.50
            citrus 50% canopy                          , 1.1,   1.5, 0.50
            citrus 20% canopy                          , 0.8,   1.1, 0.50
            conifer trees                              , 1.0,   1.5, 0.70
            kiwi                                       , 0.7,   1.3, 0.35
            olives 40% to 60% ground coverage by canopy, 1.2,   1.7, 0.65
            pistachios                                 , 1.0,   1.5, 0.40
            walnut orchard                             , 1.7,   2.4, 0.50"""
        self.table22 = pd.read_csv(io.StringIO(table22data),
                                   names=['Crop','Zrmax1','Zrmax2','pbase'],
                                   dtype=str)
        for col in self.table22.columns:
            self.table22[col] = self.table22[col].str.strip()

        self.synonyms = {
            "brussel sprouts": ["crucifers"],
            "brussel sprout": ["crucifers"],
            "red cabbage": ["cabbage","crucifers"],
            "green cabbage": ["cabbage","crucifers"],
            "purple cabbage": ["cabbage","crucifers"],
            "savoy cabbage": ["cabbage","crucifers"],
            "cabbages": ["cabbage","crucifers"],
            "elephant garlic": ["garlic"],
            "wild garlic": ["garlic"],
            "single clove garlic": ["garlic"],
            "romaine lettuce": ["lettuce"],
            "romaine": ["lettuce"],
            "onions": ["onion"],
            "red radish": ["radish"],
            "green radish": ["radish"],
            "black radish": ["radish"],
            "oilseed radish": ["radish"],
            "wild radish": ["radish"],
            "daikon": ["radish"],
            "radishes": ["radish"],
            "egg plant": ["eggplant"],
            "garden egg": ["eggplant"],
            "eggplants": ["eggplant"],
            "egg plants": ["eggplant"],
            "bell pepper": ["sweet pepper"],
            "bell peppers": ["sweet pepper"],
            "paprika": ["sweet pepper"],
            "capsicus": ["sweet pepper"],
            "capsicum": ["sweet pepper"],
            "tomatoes": ["tomato"],
            "true cantaloupe": ["cantaloupe"],
            "european cantaloupe": ["cantaloupe"],
            "american cantaloupe": ["cantaloupe"],
            "cantaloupes": ["cantaloupe"],
            "muskmelon": ["cantaloupe"],
            "rockmelon": ["cantaloupe"],
            "spanspek": ["cantaloupe"],
            "cucumbers": ["cucumber"],
            "pumpkins": ["pumpkin"],
            "winter squashes": ["winter squash"],
            "kabocha": ["winter squash"],
            "butternut squash": ["winter squash"],
            "spaghetti squash": ["winter squash"],
            "delicata squash": ["winter squash"],
            "acorn squash": ["winter squash"],
            "summer squash": ["zucchini"],
            "crookneck squash": ["squash"],
            "gem squash": ["squash"],
            "pattypan squash": ["squash"],
            "straightneck squash": ["squash"],
            "zucchinis": ["zucchini"],
            "kamokamo": ["squash"],
            "aehobak": ["zucchini"],
            "tromboncino": ["squash"],
            "zucchetta": ["squash"],
            "squashes": ["squash"],
            "melons": ["melon"],
            "watermelon": ["water melon"],
            "watermelons": ["watermelon", "water melon"],
            "water melon": ["watermelon"],
            "water melons": ["watermelon","water melon"],
            "true melon": ["sweet melons"],
            "horned melon": ["sweet melons"],
            "table beets": ["beets table"],
            "table beet": ["beets table"],
            "beetroot": ["beets table"],
            "garden beets": ["beets table"],
            "garden beet": ["beets table"],
            "dinner beets": ["beets table"],
            "dinner beet": ["beets table"],
            "parsnips": ["parsnip"],
            "potatoes": ["potato"],
            "spud": ["potato"],
            "sweet potatoes": ["sweet potato"],
            "turnips": ["turnip"],
            "rutabagas": ["rutabaga"],
            "swedish turnip": ["turnip"],
            "white turnip": ["turnip"],
            "sugarbeet": ["sugar beet"],
            "sugar beets": ["sugar beet","sugarbeet"],
            "sugarbeets": ["sugar beet","sugarbeet"],
            "green beans": ["beans green"],
            "greenbeans": ["beans green"],
            "green bean": ["beans green"],
            "greenbean": ["beans green"],
            "bean": ["chickpea","garbanzo"],
            "beans": ["bean","chickpea","garbanzo"],
            "dry beans": ["beans dry"],
            "dry broad beans": ["beans dry"],
            "pulses": ["beans dry"],
            "chick pea": ["chickpea"],
            "chickpeas": ["chickpea"],
            "chick peas": ["chickpea"],
            "fababean": ["faba bean"],
            "garbanzo bean": ["garbanzo"],
            "garbanzo beans": ["garbanzo"],
            "peanuts": ["peanut"],
            "groundnuts": ["groundnut"],
            "goober": ["peanut"],
            "goober pea": ["peanut"],
            "goober peas": ["peanut"],
            "pindar": ["peanut"],
            "monkey nut": ["peanut"],
            "lentils": ["lentil"],
            "black-eyed pea": ["cowpea"],
            "black-eyed peas": ["cowpea"],
            "soy bean": ["soybeans"],
            "soy beans": ["soybeans"],
            "soya bean": ["soybeans"],
            "artichokes": ["artichoke"],
            "french artichoke": ["artichoke"],
            "globe artichoke": ["artichoke"],
            "green artichoke": ["artichoke"],
            "mentha": ["mint"],
            "strawberry": ["strawberries"],
            "garden strawberry": ["strawberries"],
            "upland cotton": ["cotton"],
            "extra-long staple cotton": ["cotton"],
            "tree cotton": ["cotton"],
            "levant cotton": ["cotton"],
            "pima cotton": ["cotton"],
            "common flax": ["flax"],
            "linseed": ["flax"],
            "castorbean": ["castor bean"],
            "castor bean": ["castorbean"],
            "castorbeans": ["castorbean", "castor bean"],
            "castor beans": ["castorbean", "castor bean"],
            "ricinus": ["castorbean", "castor bean"],
            "castor oil plant": ["castorbean", "castor bean"],
            "rape": ["rapeseed"],
            "oilseed rape": ["rapeseed"],
            "simsim": ["sesame"],
            "benne": ["sesame"],
            "gingelly": ["sesame"],
            "sunflowers": ["sunflower"],
            "common sunflowers": ["sunflower"],
            "common sunflower": ["sunflower"],
            "common oat": ["oats"],
            "common oats": ["oats"],
            "corn": ["maize"],
            "zea mays": ["maize"],
            "field corn": ["maize field", "maize grain"],
            "sweet corn": ["maize sweet"],
            "millets": ["millet"],
            "finger millet": ["millet"],
            "pearl millet": ["millet"],
            "little millet": ["millet"],
            "sorghum": ["sorghum"],
            "great millet": ["sorghum"],
            "broomcorn": ["sorghum"],
            "guinea corn": ["sorghum"],
            "durra": ["sorghum"],
            "imphee": ["sorghum"],
            "jowar": ["sorghum"],
            "milo": ["sorghum"],
            "indica rice": ["rice"],
            "japonica rice": ["rice"],
            "lucerne": ["alfalfa"],
            "bermudagrass": ["bermuda"],
            "bermuda grass": ["bermuda"],
            "coastal bermudagrass": ["bermuda"],
            "berseem clover": ["clover"],
            "egyptian clover": ["clover"],
            "ryegrass": ["rye grass"],
            "sudan grass": ["sudan"],
            "sudan grass hay": ["sudan"],
            "turf grass": ["turfgrass"],
            "lawn grass": ["turfgrass"],
            "bananas": ["banana"],
            "cacao tree": ["cacao"],
            "cocoa tree": ["cacao"],
            "cocoa": ["cacao"],
            "arabica coffee": ["coffee"],
            "coffea arabica": ["coffee"],
            "coffee plant": ["coffee"],
            "dates": ["date palms"],
            "palms": ["palm trees"],
            "pineapples": ["pineapple"],
            "ananas": ["pineapple"],
            "rubber plant": ["rubber trees"],
            "sharinga tree": ["rubber trees"],
            "para rubber tree": ["rubber trees"],
            "rubber plants": ["rubber trees"],
            "tea tree": ["tea"],
            "tea plant": ["tea"],
            "tea shrub": ["tea"],
            "tea trees": ["tea"],
            "tea plants": ["tea"],
            "tea shrubs": ["tea"],
            "berry": ["berries", "strawberries"],
            "berry shrub": ["berries bushes"],
            "berry bushes": ["berries bushes"],
            "berry bush": ["berries bushes"],
            "blueberries": ["berries bushes"],
            "blueberry": ["berries bushes"],
            "boysenberry": ["berries bushes"],
            "elderberry": ["berries bushes"],
            "blackberry": ["berries bushes"],
            "blackberries": ["berries bushes"],
            "raspberry": ["berries bushes"],
            "raspberries": ["berries bushes"],
            "wine grape": ["grapes wine"],
            "wine grapes": ["grapes wine"],
            "table grape": ["grapes table"],
            "table grapes": ["grapes table"],
            "raisin": ["grapes table"],
            "hops plant": ["hops"],
            "almond tree": ["almonds"],
            "almond trees": ["almonds"],
            "apple tree": ["apples"],
            "apple trees": ["apples"],
            "apple": ["apples"],
            "cherry": ["cherries"],
            "cherry tree": ["cherries"],
            "cherry trees": ["cherries"],
            "pear tree": ["pears"],
            "pear trees": ["pears"],
            "apricot tree": ["apricots"],
            "apricot trees": ["apricots"],
            "peach tree": ["peaches"],
            "peach trees": ["peaches"],
            "stone fruits": ["stone fruit"],
            "avocados": ["avocado"],
            "citrus trees": ["citrus"],
            "citrus tree": ["citrus"],
            "oranges": ["citrus"],
            "orange": ["citrus"],
            "orange tree": ["citrus"],
            "orange trees": ["citrus"],
            "grapefruits": ["citrus"],
            "grapefruit": ["citrus"],
            "grapefruit trees": ["citrus"],
            "grapefruit tree": ["citrus"],
            "limes": ["citrus"],
            "lime": ["citrus"],
            "lime tree": ["citrus"],
            "lime trees": ["citrus"],
            "lemon": ["citrus"],
            "lemons": ["citrus"],
            "lemon tree": ["citrus"],
            "lemon trees": ["citrus"],
            "tangerine": ["citrus"],
            "tangerines": ["citrus"],
            "tangerine trees": ["citrus"],
            "tangerine tree": ["citrus"],
            "conifer tree": ["conifer trees"],
            "coniferous trees": ["conifer trees"],
            "coniferous tree": ["conifer trees"],
            "spruce trees": ["conifer trees"],
            "spruce tree": ["conifer trees"],
            "pine trees": ["conifer trees"],
            "pine tree": ["conifer trees"],
            "fir trees": ["conifer trees"],
            "fir tree": ["conifer trees"],
            "kiwis": ["kiwi"],
            "kiwifruit": ["kiwi"],
            "chinese gooseberry": ["kiwi"],
            "chinese gooseberries": ["kiwi"],
            "kiwifruits": ["kiwi"],
            "green olive": ["olives"],
            "black olive": ["olives"],
            "green olives": ["olives"],
            "black olives": ["olives"],
            "olive tree": ["olives"],
            "olive trees": ["olives"],
            "pistachio tree": ["pistachios"],
            "pistachio trees": ["pistachios"],
            "walnuts": ["walnut"],
            "walnut tree": ["walnut"],
            "walnut trees": ["walnut"],
            "january": ["jan"],
            "february": ["feb"],
            "march": ["mar"],
            "april": ["apr"],
            "june": ["jun"],
            "july": ["jul"],
            "august": ["aug"],
            "september": ["sep"],
            "october": ["oct"],
            "november": ["nov"],
            "december": ["dec"]
        }

    def search11(self,phrase,usesyn=True):
        """Search FAO-56 Table 11 based on the provided phrase

        Parameters
        ----------
        phrase : str
            String expression for searching FAO-56 Table 11
        usesyn : boolean, optional
            Whether or not to use the synonym dictionary (default=True)

        Returns
        -------
        table11sub : DataFrame
            Subsetted FAO-56 Table 11 based on the search phrase
        """

        phrase = phrase.lower()
        phrases = [phrase]
        if usesyn and phrase in self.synonyms.keys():
            phrases += self.synonyms[phrase]
        indices = list()
        df = self.table11
        for col in ['Crop','Plant Date','Region']:
            for phrase in phrases:
                temp =  df[df[col].str.contains(phrase)].index.values
                indices += temp.tolist()
        table11sub = self.table11.iloc[list(set(indices))]
        return table11sub

    def search12(self,phrase,usesyn=True):
        """Search FAO-56 Table 12 based on the provided phrase

        Parameters
        ----------
        phrase : str
            String expression for searching FAO-56 Table 12
        usesyn : boolean, optional
            Whether or not to use the synonym dictionary (default=True)

        Returns
        -------
        table12sub : DataFrame
            Subsetted FAO-56 Table 12 based on the search phrase
        """

        phrase = phrase.lower()
        phrases = [phrase]
        if usesyn and phrase in self.synonyms.keys():
            phrases += self.synonyms[phrase]
        indices = list()
        df = self.table12
        for phrase in phrases:
            temp = df[df['Crop'].str.contains(phrase)].index.values
            indices += temp.tolist()
        table12sub = self.table12.iloc[list(set(indices))]
        return table12sub

    def search17(self,phrase,usesyn=True):
        """Search FAO-56 Table 17 based on the provided phrase

        Parameters
        ----------
        phrase : str
            String expression for searching FAO-56 Table 17
        usesyn : boolean, optional
            Whether or not to use the synonym dictionary (default=True)

        Returns
        -------
        table17sub : DataFrame
            Subsetted FAO-56 Table 17 based on the search phrase
        """

        phrase = phrase.lower()
        phrases = [phrase]
        if usesyn and phrase in self.synonyms.keys():
            phrases += self.synonyms[phrase]
        indices = list()
        df = self.table17
        for phrase in phrases:
            temp = df[df['Crop'].str.contains(phrase)].index.values
            indices += temp.tolist()
        table17sub = self.table17.iloc[list(set(indices))]
        return table17sub

    def search22(self,phrase,usesyn=True):
        """Search FAO-56 Table 22 based on the provided phrase

        Parameters
        ----------
        phrase : str
            String expression for searching FAO-56 Table 22
        usesyn : boolean, optional
            Whether or not to use the synonym dictionary (default=True)

        Returns
        -------
        table17sub : DataFrame
            Subsetted FAO-56 Table 22 based on the search phrase
        """

        phrase = phrase.lower()
        phrases = [phrase]
        if usesyn and phrase in self.synonyms.keys():
            phrases += self.synonyms[phrase]
        indices = list()
        df = self.table22
        for phrase in phrases:
            temp = df[df['Crop'].str.contains(phrase)].index.values
            indices += temp.tolist()
        table22sub = self.table22.iloc[list(set(indices))]
        return table22sub
